/*======================================================================*/
/* Excel.cpp: CExcel Excel automation object implementation				*/
/* Copyright 2001-2003, (C) Jean-Pierre Menicucci - all rights reserved	*/
/*======================================================================*/

#include "stdafx.h"						// base includes
#include "Bstr.h"						// CBSTR BSTR wrapper class
#include "Lang.h"						// CLang resource handling class
#include "Genfun.h"						// general functions includes
#include "Excel.h"						// CExcel Excel automation class

/*======================================================================*/
/* Close : release dispatch interface									*/
/*======================================================================*/

HRESULT CExcel::Close (
	void
	)
{
	/*******************/
	/* workbooks.close */
	/*******************/
	{
	VARIANT aw = Get (m_pDispApp, L"ActiveWorkbook");		// .ActiveWorkbook
	Call (aw.pdispVal, L"Close", dpNoArgs(), NULL);			// .Close
	aw.pdispVal->Release();									// release interface
	}

	/******************************/
	/* release dispatch interface */
	/******************************/

	if (m_pDispApp) {
		m_pDispApp->Release();								// release interface
		m_pDispApp = 0;
		}

	return 0;
}

/*======================================================================*/
/* CreateChart : create chart and copy to clipboard						*/
/*======================================================================*/

HRESULT CExcel::CreateChart (
	LPSTR lpTit, int nTitLen,			// graph title
	LPSTR lpXax, int nXaxLen,			// X axis title
	LPSTR lpYax, int nYaxLen,			// Y axis title
	LPSTR lpDa2, int nDa2Len,			// graph data
	LPSTR lpYmi, int nYmiLen,			// Y axis min
	LPSTR lpYmx, int nYmxLen,			// Y axis max
	int nChartType						// chart type
	)
{
	int nRow, nCol;						// dimensions

	FillSheet (lpDa2, nDa2Len, &nRow, &nCol);

	/******************************/
	/* .ActiveWorkbook.Charts.Add */
	/******************************/

	VARIANT awb = Get (m_pDispApp, L"ActiveWorkbook");		// .ActiveWorkbook

	VARIANT hs = Get (awb.pdispVal, L"Charts");				// .Charts

	Call (hs.pdispVal, L"Add", dpNoArgs(), 0);				// .Add

	/*****************************/
	/* .ActiveChart.Name="Graph" */
	/*****************************/

	VARIANT ach = Get (m_pDispApp, L"ActiveChart");			// .ActiveChart

	PutBstr (ach.pdispVal, L"Name", L"Graph");				// .Name="Graph"

	/*************************************/
	/* .ActiveChart.ChartType=nChartType */
	/*************************************/

	PutI4 (ach.pdispVal, L"ChartType", nChartType);			// .ChartType=nChartType

	/***********************************************************************************/
	/* .SetSourceData Source:=.Range(.Cells(1,1),.Cells(nRow,nCol)), PlotBy:=xlColumns */
	/***********************************************************************************/

	VARIANT shs = Get (m_pDispApp, L"Sheets");				// .Sheets

	VARIANT x[2];
	x[0].vt = VT_BSTR;
	x[0].bstrVal = CBSTR ("Data");
	DISPPARAMS dpx = { x, 0, 1, 0 };

	VARIANT sh = Get (shs.pdispVal, L"Item", &dpx);			// .Sheets("Data")

	x[1].vt = VT_I4;
	x[1].lVal = 1;
	x[0].vt = VT_I4;
	x[0].lVal = 1;
	dpx.cArgs = 2;

	VARIANT lt = Get (sh.pdispVal, L"Cells", &dpx);			// .Cells(1,1)

	x[1].lVal = nRow;
	x[0].lVal = nCol;

	VARIANT rb = Get (sh.pdispVal, L"Cells", &dpx);			// .Cells(nRow,nCol)

	x[1].vt = VT_DISPATCH;
	x[1].pdispVal = lt.pdispVal;
	x[0].vt = VT_DISPATCH;
	x[0].pdispVal = rb.pdispVal;

	VARIANT rg = Get (sh.pdispVal, L"Range", &dpx);			// .Range(.Cells(1,1),.Cells(nRow,nCol))

	x[1].pdispVal = rg.pdispVal;
	x[0].vt = VT_I4;
	x[0].lVal = 2;

	Call (ach.pdispVal, L"SetSourceData", &dpx, 0);			// .SetSourceData Source:=r, PlotBy:=xlColumns

	/*==================================================================*/
	/* graph title														*/
	/*==================================================================*/

	if (nTitLen) {

		/******************************/
		/* .ActiveChart.HasTitle=true */
		/******************************/

		PutBool (ach.pdispVal, L"HasTitle", TRUE);			// .HasTitle=true

		/**************************************/
		/* .ActiveChart.ChartTitle.Text=lpTit */
		/**************************************/

		VARIANT ct = Get (ach.pdispVal, L"ChartTitle");		// .ChartTitle

		CBSTR bsTit (nTitLen, lpTit);

		PutBstr (ct.pdispVal, L"Text", bsTit);				// .Text=lpTit

		/******************************************/
		/* .ActiveChart.ChartTitle.Font.Bold=true */
		/******************************************/

		VARIANT f = Get (ct.pdispVal, L"Font");				// .Font

		PutI4 (f.pdispVal, L"Size", 18);					// .Size=18

		PutBool (f.pdispVal, L"Bold", TRUE);				// .Bold=true

		f.pdispVal->Release();								// release interface
		ct.pdispVal->Release();								// release interface
		}

	/*****************************************************************/
	/* .ActiveChart.ChartArea.Border.Weight=xlHairLine (thin border) */
	/*****************************************************************/

	VARIANT cha = Get (ach.pdispVal, L"ChartArea");			// .ChartArea

	VARIANT b = Get (cha.pdispVal, L"Border");				// .Border

	PutI4 (b.pdispVal, L"Weight", 1);						// .Weight=xlHairLine

	/*****************************************************/
	/* .ActiveChart.PlotArea.Border.Interior.Color=white */
	/*****************************************************/

	VARIANT pa = Get (ach.pdispVal, L"PlotArea");			// .PlotArea

	VARIANT pab = Get (pa.pdispVal, L"Border");				// .Border

	VARIANT i = Get (pa.pdispVal, L"Interior");				// .Interior

	PutI4 (i.pdispVal, L"Color", RGB (255,255,255));		// .Color=white

	switch (nChartType) {
	case 51:							// xlColumnClustered
	case 54:							// xl3DColumnClustered
	case 65:							// xlLineMarkers
	case -4101:							// xl3DLine
		{
		/***********************************************************/
		/* .ActiveChart.Axes(xlCategory).TickLabels.Orientation=45 */
		/***********************************************************/

		x[0].vt = VT_I4;
		x[0].lVal = 1;										// xlCategory
		dpx.cArgs = 1;

		VARIANT xax = Get (ach.pdispVal, L"Axes", &dpx);	// .Axes(xlCategory)

		VARIANT xtl = Get (xax.pdispVal, L"TickLabels");	// .TickLabels

		PutI4 (xtl.pdispVal, L"Orientation", 45);			// .Orientation=45

		/*********************************************************/
		/* .ActiveChart.Axes(xlCategory).TickLabels.Font.Size=12 */
		/*********************************************************/
		{
		VARIANT f = Get (xtl.pdispVal, L"Font");			// .Font

		PutI4 (f.pdispVal, L"Size", 12);					// .Size=12

		f.pdispVal->Release();								// release interface
		}

		/*==============================================================*/
		/* x axis title													*/
		/*==============================================================*/

		if (nXaxLen) {

			/*************************************************/
			/* .ActiveChart.Axes(xlCategory).HasTitle = True */
			/*************************************************/

			PutBool (xax.pdispVal, L"HasTitle", TRUE);		// .HasTitle=true

			/******************************************************/
			/* .ActiveChart.Axes(xlCategory).AxisTitle.Text=lpXax */
			/******************************************************/

			VARIANT xat = Get (xax.pdispVal, L"AxisTitle");	// .AxisTitle

			CBSTR bsXax (nXaxLen, lpXax);

			PutBstr (xat.pdispVal, L"Text", bsXax);			// .Text=lpXax

			/********************************************************/
			/* .ActiveChart.Axes(xlCategory).AxisTitle.Font.Size=14 */
			/********************************************************/

			VARIANT f = Get (xat.pdispVal, L"Font");		// .Font

			PutI4 (f.pdispVal, L"Size", 14);				// .Size=14

			f.pdispVal->Release();							// release interface
			xat.pdispVal->Release();						// release interface
			}

		x[0].lVal = 2;										// xlValue

		VARIANT yax = Get (ach.pdispVal, L"Axes", &dpx);	// .Axes(xlVale)

		/*==============================================================*/
		/* y axis title													*/
		/*==============================================================*/

		if (nYaxLen) {

			/**********************************************/
			/* .ActiveChart.Axes(xlValue).HasTitle = True */
			/**********************************************/

			PutBool (yax.pdispVal, L"HasTitle", TRUE);		// .HasTitle=true

			/***************************************************/
			/* .ActiveChart.Axes(xlValue).AxisTitle.Text=lpYax */
			/***************************************************/

			VARIANT yat = Get (yax.pdispVal, L"AxisTitle");	// .AxisTitle

			CBSTR bsYax (nYaxLen, lpYax);

			PutBstr (yat.pdispVal, L"Text", bsYax);			// .Text=lpYax

			/*****************************************************/
			/* .ActiveChart.Axes(xlValue).AxisTitle.Font.Size=14 */
			/*****************************************************/

			VARIANT f = Get (yat.pdispVal, L"Font");		// .Font

			PutI4 (f.pdispVal, L"Size", 14);				// .Size=14

			f.pdispVal->Release();
			yat.pdispVal->Release();

			/*************************************************/
			/* .ActiveChart.Axes(xlValue).MinimumScale=lpYmi */
			/*************************************************/

			if (lpYmi) {
				double d = atof (lpYmi);
				PutDouble (yax.pdispVal, L"MinimumScale", d); // .MinimumScale=lpYmi
				}

			/*************************************************/
			/* .ActiveChart.Axes(xlValue).MinimumScale=lpYmx */
			/*************************************************/

			if (lpYmi) {
				double d = atof (lpYmx);
				PutDouble (yax.pdispVal, L"MaximumScale", d); // .MaximumScale=lpYmx
				}
			}

		/**************************************************************/
		/* .ActiveChart.Axes(xlValue).TickLabels.NumberFormat="# ##0" */
		/**************************************************************/

		VARIANT ytl = Get (yax.pdispVal, L"TickLabels");	// .TickLabels

		PutBstr (ytl.pdispVal,L"NumberFormat",L"# ##0");	// .NumberFormat="# ##0"

		/******************************************************/
		/* .ActiveChart.Axes(xlValue).TickLabels.Font.Size=12 */
		/******************************************************/
		{
		VARIANT f = Get (ytl.pdispVal, L"Font");			// .Font

		PutI4 (f.pdispVal, L"Size", 12);					// .Size=12

		f.pdispVal->Release();								// release interface
		}
		/********************************************/
		/* .ActiveChart.Axes(xlValue).MajorUnit=1.0 */
		/********************************************/

		VARIANT mu = Get (yax.pdispVal, L"MajorUnit");		// .MajorUnit

		if (mu.dblVal < 1.0)
			PutDouble (yax.pdispVal, L"MajorUnit", 1.0);	// .MajorUnit=1.0

		/************************************/
		/* .ActiveChart.Legeng.Font.Size=12 */
		/************************************/
		{
		VARIANT l = Get (ach.pdispVal, L"Legend");			// .Legend

		VARIANT f = Get (l.pdispVal, L"Font");				// .Font

		PutI4 (f.pdispVal, L"Size", 12);					// .Size=12

		f.pdispVal->Release();								// release interface
		}
		/**********************/
		/* release interfaces */
		/**********************/

		ytl.pdispVal->Release();							// release interface
		yax.pdispVal->Release();							// release interface
		xtl.pdispVal->Release();							// release interface
		xax.pdispVal->Release();							// release interface
		}

		/********************************************/
		/* .ActiveChart.PlotArea.Border.Color=black */
		/********************************************/

		PutI4 (pab.pdispVal, L"Color", RGB (0,0,0));		// .Color=black

		break;

	case 5:								// xlPie
	case -4102:							// xl3DPie
		{
		/**********************************************************************/
		/* .ActiveChart.ApplyDataLabels Type:=xlDataLabelsShowLabelAndPercent */
		/**********************************************************************/

		x[0].vt = VT_I4;
		x[0].lVal = 5;										// xlDataLabelsShowLabelAndPercent
		dpx.cArgs = 1;

		Call (ach.pdispVal, L"ApplyDataLabels", &dpx, 0);	// .ApplyDataLabels Type:=xlDataLabelsShowLabelAndPercent

		/*****************************************************/
		/* .ActiveChart.PlotArea.Border.LineStyle=xlLineNone */
		/*****************************************************/

		PutI4 (pab.pdispVal, L"LineStyle", -4142);			// .LineStyle=xlLineNone
		}
		break;
		}

	switch (nChartType) {
	case 65:							// xlLineMarkers
	case -4101:							// xl3DLine
		{
		/************************/
		/* set up series colors */
		/************************/

		VARIANT as = Get (m_pDispApp, L"ActiveSheet");		// .ActiveSheet

		VARIANT sc = Get (as.pdispVal,L"SeriesCollection");	// .SeriesCollection

		VARIANT c = Get (sc.pdispVal, L"Count");			// .Count

		switch (min (5, c.lVal)) {
		case 5:
			SeriesColor (sc, 5, 12);						// .SeriesCollection(5).ColorIndex=12
		case 4:
			SeriesColor (sc, 4, 9);							// .SeriesCollection(4).ColorIndex=9
		case 3:
			SeriesColor (sc, 3, 10);						// .SeriesCollection(3).ColorIndex=10
			}

		/**********************/
		/* release interfaces */
		/**********************/

		sc.pdispVal->Release();								// release interface
		as.pdispVal->Release();								// release interface
		}
		break;
		}

	/*********************************/
	/* .ActiveChart.ChartArea.Select */
	/*********************************/

	Call (cha.pdispVal, L"Select", dpNoArgs(), 0);			// .Select

	/*******************************/
	/* .ActiveChart.ChartArea.Copy */
	/*******************************/

	Call (cha.pdispVal, L"Copy", dpNoArgs(), 0);			// .Copy

	/**********************/
	/* release interfaces */
	/**********************/

	i.pdispVal->Release();									// release interface
	pab.pdispVal->Release();								// release interface
	pa.pdispVal->Release();									// release interface
	b.pdispVal->Release();									// release interface
	cha.pdispVal->Release();								// release interface
	rg.pdispVal->Release();									// release interface
	rb.pdispVal->Release();									// release interface
	lt.pdispVal->Release();									// release interface
	sh.pdispVal->Release();									// release interface
	shs.pdispVal->Release();								// release interface
	ach.pdispVal->Release();								// release interface
	hs.pdispVal->Release();									// release interface
	awb.pdispVal->Release();								// release interface

	return 0;
}

/*======================================================================*/
/* DeleteChart : delete chart sheet										*/
/*======================================================================*/

HRESULT CExcel::DeleteChart (
	void								// no argument
	)
{
	/***************************************/
	/* .ActiveWindow.SelectedSheets.Delete */
	/***************************************/

	VARIANT aw = Get (m_pDispApp, L"ActiveWindow");			// .ActiveWindow

	VARIANT ssh = Get (aw.pdispVal, L"SelectedSheets");		// .SelectedSheets

	Call (ssh.pdispVal, L"Delete", dpNoArgs(), 0);			// .Delete

	/****************/
	/* .Cells.Clear */
	/****************/

	VARIANT ces = Get (m_pDispApp, L"Cells");				// .Cells

	Call (ces.pdispVal, L"Clear", dpNoArgs(), 0);			// .Clear

	/**********************/
	/* release interfaces */
	/**********************/

	ces.pdispVal->Release();								// release interface
	ssh.pdispVal->Release();								// release interface
	aw.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* FillSheet : insert data in worksheet									*/
/*======================================================================*/

HRESULT CExcel::FillSheet (
	LPSTR lpDat,						// line buffer
	int nDat,							// line length
	int* pnRow,							// row#
	int* pnCol							// col#
	)
{
	VARIANT x[2];						// .Cells call data
	x[0].vt = VT_I4;					// column number
	x[1].vt = VT_I4;					// row number
	DISPPARAMS dp = { x, 0, 2, 0 };

	VARIANT y[1];						// .Value call data
	y[0].vt = VT_BSTR;					// text value
	DISPPARAMS dq = v2dp (y);

	VARIANT as = Get (m_pDispApp, L"ActiveSheet");			// Range, .ActiveSheet

	/**************************/
	/* for all lines in table */
	/**************************/

	for (*pnRow = 0; nDat > 0; (*pnRow)++) {				// line number

		for (int iDat = 0; iDat < nDat; iDat++)				// get line length
			if (*(lpDat+iDat) == '\x0A')					// line-feed-delimited
				break;

		LPSTR lpCel = lpDat;			// cell buffer
		int nCel = iDat;				// cell length

		/***************************/
		/* for all columns in line */
		/***************************/

		for (*pnCol = 0; nCel > 0; (*pnCol)++) {			// column number

			for (int iCel = 0; iCel < nCel; iCel++)	{		// get cell length
				if (*(lpCel+iCel) == '.')					// replace dots
					*(lpCel+iCel) = ',';					// by colons
				else if (*(lpCel+iCel) == '\x09')			// tab-delimited
					break;
				}

			/************/
			/* get cell */
			/************/

			x[1].lVal = *pnRow+1;
			x[0].lVal = *pnCol+1;

			VARIANT rg = Get (as.pdispVal, L"Cells", &dp);	// .Cells(*pnRow+1,*pnCol+1)

			/*****************************************************/
			/* if 1st column and date, reformat and set up value */
			/*****************************************************/

			if (*pnCol == 0 && iCel > 7 && *(lpCel+2) == '/' && *(lpCel+5) == '/') {

				int i = iCel+2;								// new length

				LPSTR lp = (LPSTR) malloc (i);				// allocate storage
				LPSTR lpM = CLang::MonthName (lpCel+3);		// allocate month name

				nCpy (lp, lpCel, 2);						// reformat
				*(lp+2) = '-';
				nCpy (lp+3, lpM, 4);
				*(lp+7) = '-';
				nCpy (lp+8, lpCel+6, iCel-6);

				CBSTR bsCel (i, lp);
				y[0].bstrVal = bsCel;
				Put (rg.pdispVal, L"Value", &dq, 0);		// .Value=lpCel

				free (lpM);									// free month name
				free (lp);									// free storage
				}

			/**************************/
			/* else just set up value */
			/**************************/

			else {
				CBSTR bsCel (iCel, lpCel);
				y[0].bstrVal = bsCel;
				Put (rg.pdispVal, L"Value", &dq, 0);		// .Value=lpCel
				}

			/**********************/
			/* release interfaces */
			/**********************/

			rg.pdispVal->Release();							// release interface

			lpCel += iCel+1;								// move to next cell
			nCel -= iCel+1;
			}

		lpDat += iDat+1;									// move to next line
		nDat -= iDat+1;
		}

	/**********************/
	/* release interfaces */
	/**********************/

	as.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* Open : open Excel instance											*/
/*======================================================================*/

HRESULT CExcel::Open (
	void
	)
{
	/*****************************************/
	/* start Excel and get IDispatch pointer */
	/*****************************************/

	CLSID clsid;						// Excel server clsid

	HRESULT hr = CLSIDFromProgID (L"Excel.Application", &clsid);

	if (FAILED (hr)) {
		CLang::SysMesBox (hr);
		return hr;
		}

	IUnknown* pUnk;
	hr = GetActiveObject (clsid, NULL, (IUnknown**) &pUnk);
	BOOL bNew = FAILED (hr);

	if (FAILED (hr))
		hr = CoCreateInstance (clsid, 0, CLSCTX_LOCAL_SERVER, IID_IDispatch, (void **) &m_pDispApp);
	else
		hr = pUnk->QueryInterface (IID_IDispatch, (void**) &m_pDispApp);

	if (FAILED (hr)) {
		CLang::SysMesBox (hr);
		return hr;
		}

	/*********************/
	/* get excel version */
	/*********************/

	VARIANT v;
	Get (m_pDispApp, L"Version", dpNoArgs(), &v);			// .Version
	int n = SysStringLen (v.bstrVal);
	LPSTR lp = (LPSTR) malloc (n);
	WideCharToMultiByte (CP_ACP, 0, v.bstrVal, n, lp, n, 0, 0);
	m_nVersion = atoi (lp);
	free (lp);

	/*************************/
	/* DisplayAlerts = false */
	/*************************/

	PutBool (m_pDispApp, L"DisplayAlerts", FALSE);			// DisplayAlerts = false

	/*******************/
	/*  workbooks.add  */
	/*******************/
	{
	VARIANT w = Get (m_pDispApp, L"Workbooks");				// Workbooks
	Call (w.pdispVal, L"Add", dpNoArgs(), NULL);			// .Add
	w.pdispVal->Release();									// release interface
	}
	/**********************/
	/* activesheet.delete */
	/**********************/
	{
	VARIANT as = Get (m_pDispApp, L"ActiveSheet");			// ActiveSheet
	Call (as.pdispVal, L"Delete", dpNoArgs(), NULL);		// .Delete
	as.pdispVal->Release();									// release interface
	}
	/**********************/
	/* activesheet.delete */
	/**********************/
	{
	VARIANT as = Get (m_pDispApp, L"ActiveSheet");			// ActiveSheet
	Call (as.pdispVal, L"Delete", dpNoArgs(), NULL);		// .Delete
	as.pdispVal->Release();									// release interface
	}
	/*****************************/
	/* activesheet.Name = "Data" */
	/*****************************/
	{
	VARIANT as = Get (m_pDispApp, L"ActiveSheet");			// ActiveSheet

	BSTR bs = SysAllocString (L"Data");
	VARIANT w[1];
	w[0].vt = VT_BSTR;
	w[0].bstrVal = bs;
	DISPPARAMS dp = v2dp (w);

	Put (as.pdispVal, L"Name", &dp, NULL);					// .Name = "Data"

	SysFreeString (bs);

	as.pdispVal->Release();									// release interface
	}
	/*******************************************/
	/* Excel.WindowState = wdWindowStateNormal */
	/*******************************************/

	PutI4 (m_pDispApp, L"WindowState", 0xFFFFEFD1);			// WindowState = xlNormal

	/************************/
	/* excel.visible = true */
	/************************/

	PutBool (m_pDispApp, L"Visible", TRUE);					// Visible = true

	return hr;
}

/*======================================================================*/
/* SeriesColor : set up series color									*/
/*======================================================================*/

HRESULT CExcel::SeriesColor (
	VARIANT sc,							// series collection
	int si,								// series index
	int ci								// color index
	)
{
	VARIANT c;
	c.vt = VT_I4;
	c.lVal = si;
	DISPPARAMS dpc = { &c, 0, 1, 0 };

	VARIANT s = Call (sc.pdispVal, L"Item", &dpc);			// .SeriesCollection(si)

	VARIANT b = Get (s.pdispVal, L"Border");				// .Border

	PutI4 (b.pdispVal, L"ColorIndex", ci);					// .Border.ColorIndex=ci
	PutI4 (s.pdispVal, L"MarkerBackgroundColorIndex", ci);	// .MarkerBackgroundColorIndex=ci
	PutI4 (s.pdispVal, L"MarkerForegroundColorIndex", ci);	// .MarkerForegroundColorIndex=ci

	/**********************/
	/* release interfaces */
	/**********************/

	b.pdispVal->Release();									// release interface
	s.pdispVal->Release();									// release interface

	return 0;
}
