/*======================================================================*/
/* Genfun.h: general functions											*/
/* Copyright 2001-2003, (C) Jean-Pierre Menicucci - all rights reserved	*/
/*======================================================================*/

#ifndef __Genfun_H_
#define __Genfun_H_

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*//////////////////////////////////////////////////////////////////////*/
/* string handling functions											*/
/*//////////////////////////////////////////////////////////////////////*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/*======================================================================*/
/* Upper: set a character to uppercase									*/
/* return:																*/
/*		uppercase character												*/
/*======================================================================*/

unsigned char Upper (
	unsigned char c						// character to set as uppercase
	);

/*======================================================================*/
/* Lower: set a character to lowercase									*/
/* return:																*/
/*		lowercase character												*/
/*======================================================================*/

unsigned char Lower (
	unsigned char c						// charater to set as lowercase
	);

/*======================================================================*/
/* Upper: set a string to uppercase										*/
/* return:																*/
/*		void															*/
/*======================================================================*/

void Upper (
	LPTSTR lp,							// string to uppercase
	int n								// string length
	);

/*======================================================================*/
/* Lower: set a string to lowercase										*/
/* return:																*/
/*		void															*/
/*======================================================================*/

void Lower (
	LPTSTR lp,							// string to lowercase
	int n								// string length
	);

/*======================================================================*/
/* nCpy: copying a string to another									*/
/* return:																*/
/*		string length													*/
/*======================================================================*/

int __inline nCpy (
	LPTSTR lpOut,						// output string
	LPCTSTR lpIn,						// input string
	int nIn								// copy length
	)
{
	for (int n = 0; n < nIn; n++)		// copy
		*(lpOut++) = *(lpIn++);

	return nIn;							// length
}

/*======================================================================*/
/* nCat: concatenating two strings										*/
/* return:																*/
/*		string length after concatenation								*/
/*======================================================================*/

int __inline nCat (
	LPTSTR lpOut,						// output string
	int nOffset,						// initial length
	LPCTSTR lpIn,						// input string
	int nIn								// length to copy
	)
{
	lpOut += nOffset;					// initial length

	for (int n = 0; n < nIn; n++)
		*(lpOut++) = *(lpIn++);			// copy

	return (nOffset + nIn);				// new length
}

/*======================================================================*/
/* nCmp: comparing two strings											*/
/* return:																*/
/*		-1		string 1 < string 2										*/
/*		0		string 1 == string 2									*/
/*		1		string 1 > string 2										*/
/*======================================================================*/

int __inline nCmp (
	LPCTSTR lpIn1,						// string 1
	LPCTSTR lpIn2,						// string 2
	int nIn								// length to compare
	)
{
	for (int n = 0; n < nIn; lpIn1++, lpIn2++, n++) {
		if (*lpIn1 < *lpIn2)
			return -1;					// lower
		if (*lpIn1 > *lpIn2)
			return 1;					// higher
		}
	return 0;							// equal
}

/*======================================================================*/
/* niCmp: comparing two strings ignoring case							*/
/* return:																*/
/*		-1		string 1 < string 2										*/
/*		0		string 1 == string 2									*/
/*		1		string 1 > string 2										*/
/*======================================================================*/

int __inline niCmp (
	LPCTSTR lpIn1,						// string 1
	LPCTSTR lpIn2,						// string 2
	int nIn								// length to compare
	)
{
	for (int n = 0; n < nIn; lpIn1++, lpIn2++, n++) {
		if (Upper (*lpIn1) < Upper (*lpIn2))
			return -1;					// lower
		if (Upper (*lpIn1) > Upper (*lpIn2))
			return 1;					// higher
		}
	return 0;							// equal
}

/*======================================================================*/
/* nmCpy: copying a string to another									*/
/* return:																*/
/*		string length													*/
/*======================================================================*/

int __inline nmCpy (
	LPTSTR lpOut,						// output string
	int mOut,							// output string size
	LPCTSTR lpIn,						// input string
	int nIn								// length to copy
	)
{
	mOut--;								// end-of-string character
	if (mOut < 0)						// negative or null sized string
		return 0;

	for (int n = 0; n < nIn && n < mOut; n++)
		*(lpOut++) = *(lpIn++);			// copy

	*lpOut = '\0';						// end-of-string character

	return n;							// length
}

/*======================================================================*/
/* nmCat: concatenating two strings										*/
/* return:																*/
/*		string length after concatenation								*/
/*======================================================================*/

int __inline nmCat (
	LPTSTR lpOut,						// output string
	int mOut,							// output string size
	int nOffset,						// initial length
	LPCTSTR lpIn,						// input string
	int nIn								// length to copy
	)
{
	mOut -= nOffset - 1;				// length - offset
	if (mOut < 0)						// negative or null sized string
		return nOffset;

	lpOut += nOffset;					// pointer + offset

	for (int n = 0; n < nIn && n < mOut; n++)
		*(lpOut++) = *(lpIn++);			// copy

	*lpOut = '\0';						// end-of-string character

	return (nOffset + n);				// new length
}

/*======================================================================*/
/* nmCmp: comparing two strings											*/
/* return:																*/
/*		-1		string 1 < string 2										*/
/*		0		string 1 == string 2									*/
/*		1		string 1 > string 2										*/
/*======================================================================*/

int __inline nmCmp (
	LPCTSTR lpIn1,						// string 1
	int mIn1,							// length of lpIn1
	LPCTSTR lpIn2,						// string 2
	int nIn								// length to compare
	)
{
	if (mIn1 < nIn)						// lpIn1 shorter than lpIn2
		return -1;						// lower

	for (int n = 0; n < nIn; lpIn1++, lpIn2++, n++) {
		if (*lpIn1 < *lpIn2)
			return -1;					// lower
		if (*lpIn1 > *lpIn2)
			return 1;					// higher
		}

	if (mIn1 > nIn)						// lpIn1 longer than lpIn2
		return 1;						// higher

	return 0;							// equal
}

/*======================================================================*/
/* nmiCmp: comparing two strings ignoring case							*/
/* return:																*/
/*		-1		string 1 < string 2										*/
/*		0		string 1 == string 2									*/
/*		1		string 1 > string 2										*/
/*======================================================================*/

int __inline nmiCmp (
	LPCTSTR lpIn1,						// string 1
	int mIn1,							// length of lpIn1
	LPCTSTR lpIn2,						// string 2
	int nIn								// length to compare
	)
{
	if (mIn1 < nIn)						// lpIn1 shorter than lpIn2
		return -1;						// lower

	for (int n = 0; n < nIn; lpIn1++, lpIn2++, n++) {
		if (Upper (*lpIn1) < Upper (*lpIn2))
			return -1;					// lower
		if (Upper (*lpIn1) > Upper (*lpIn2))
			return 1;					// higher
		}

	if (mIn1 > nIn)						// lpIn1 longer than lpIn2
		return 1;						// higher

	return 0;							// equal
}

/*======================================================================*/
/* Chr: search for a character											*/
/* return:																*/
/*		character address in string										*/
/*======================================================================*/

LPTSTR __inline Chr (
	LPCTSTR lp,							// string
	TCHAR c								// character
	)
{
	LPTSTR lq;
	for (; lp && *lp; lp++)
		if (*lp == c) {
			lq = (LPTSTR) lp;
			return lq;
			}
	return 0;
}

/*======================================================================*/
/* rChr: reverse search for a character									*/
/* return:																*/
/*		character address in string										*/
/*======================================================================*/

LPTSTR __inline rChr (
	LPCTSTR lp,							// string
	TCHAR c								// character
	)
{
	LPTSTR lq = 0;
	for (; lp && *lp; lp++)
		if (*lp == c)
			lq = (LPTSTR) lp;
	return lq;
}

/*======================================================================*/
/* nLfs : get string line-feed count									*/
/* return :																*/
/*		0 - no Lf in string												*/
/*		1 - 1 Lf in string												*/
/*		2 - more than 1 Lf in string									*/
/*======================================================================*/

int __inline nLfs (
	LPSTR lp,							// string
	int n								// length
	)
{
	int nLfs = 0;						// lf count

	for (int i = 0; i < n; i++) {		// for characters in string
		if (*(lp+i) == '\x0a') {		// if line feed
			if (++nLfs > 1)				// count it
				break;
			}
		}

	return nLfs;						// lf count
}

/*======================================================================*/
/* nSpaces : get string spaces count									*/
/* return :																*/
/*		number of spaces in string										*/
/*======================================================================*/

int __inline nSpaces (
	LPSTR lp,							// string
	int n								// length
	)
{
	int nSpaces = 0;					// spaces count

	for (int i = 0; i < n; i++) {		// for characters in string
		if (*(lp+i) == ' ')				// if space
			nSpaces++;
		}

	return nSpaces;						// spaces count
}

/*======================================================================*/
/* nTabs : get string tab count											*/
/* return :																*/
/*		number of tabs in string										*/
/*======================================================================*/

int __inline nTabs (
	LPSTR lp,							// string
	int n								// length
	)
{
	int nTabs = 0;						// tab count

	for (int i = 0; i < n; i++) {		// for characters in string
		if (*(lp+i) == '\x09')			// if tab
			nTabs++;
		}

	return nTabs;						// tab count
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*//////////////////////////////////////////////////////////////////////*/
/* File functions														*/
/*//////////////////////////////////////////////////////////////////////*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/*======================================================================*/
/* FileLength: getting the size of a file								*/
/* return:																*/
/*		file size														*/
/*======================================================================*/

long FileLength (
	HFILE hFile								// file handle
	);

/*======================================================================*/
/* FileRead: reading a file into a buffer								*/
/* return:																*/
/*		0					buffer allocated and file read				*/
/*		IDS_ERROPEN			could not open file							*/
/*		IDS_ERRALLOC		could not allocate memory					*/
/*		IDS_ERRREAD			could not read file							*/
/*======================================================================*/
/* FileRead allocates memory and stores the file into it				*/
/*======================================================================*/

int FileRead (
	LPCSTR lpszFilePath,				// file path
	LPSTR *lpBuf,						// buffer pointer
	DWORD *dwBuf						// buffer length
	);

/*======================================================================*/
/* CreatePath: creating a directory path								*/
/* return:																*/
/*		0		ok														*/
/*		#0		GetLastError()											*/
/*======================================================================*/

DWORD CreatePath (
	LPTSTR lpszDirPath					// file path
	);

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*//////////////////////////////////////////////////////////////////////*/
/* OLE functions														*/
/*//////////////////////////////////////////////////////////////////////*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

static DISPID dispidNamedArgs = DISPID_PROPERTYPUT;
static DISPPARAMS dispparamsNoArgs = { NULL, NULL, 0, 0 };

/*======================================================================*/
/* dpNoArgs: return "NoArguments" DISPPARAMS structure address			*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline DISPPARAMS *dpNoArgs (
	void								// no argument
	)
{
	return &dispparamsNoArgs;
}

/*======================================================================*/
/* v2dp: setup DISPPARAMS structure										*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline DISPPARAMS v2dp (
	VARIANT *pv							// variant array
	)
{
	DISPPARAMS dp = { pv, &dispidNamedArgs, 1, 1 };

	return dp;
}

/*======================================================================*/
/* Invoke: invoke method / property via IDispatch interface				*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT Invoke (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	WORD wFlags,						// flags
	DISPPARAMS *pParams,				// parameters
	VARIANT *pRes						// result
	)
{
	DISPID dispID;						// identifier

	pDisp->GetIDsOfNames (
		IID_NULL, &pName, 1, LOCALE_USER_DEFAULT, &dispID);

	EXCEPINFO ExcepInfo;

	HRESULT hr = pDisp->Invoke (
		dispID, IID_NULL, LOCALE_SYSTEM_DEFAULT,
		wFlags, pParams, pRes, &ExcepInfo, NULL);

	if (hr)
		CLang::SysMesBox (hr);

	return hr;
}

/*======================================================================*/
/* Call: call method via IDispatch interface							*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT Call (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	DISPPARAMS *pParams,				// parameters
	VARIANT *pRes						// result
	)
{
	return Invoke (pDisp, pName, DISPATCH_METHOD, pParams, pRes);
}

/*======================================================================*/
/* Call: call method via IDispatch interface							*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline VARIANT Call (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	DISPPARAMS *pParams					// parameters
	)
{
	VARIANT v;
	
	HRESULT hr = Invoke (pDisp, pName, DISPATCH_METHOD, pParams, &v);

	return v;
}

/*======================================================================*/
/* Call: call method via IDispatch interface							*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline VARIANT Call (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName					// method / property name
	)
{
	VARIANT v;
	
	HRESULT hr = Invoke (pDisp, pName, DISPATCH_METHOD, dpNoArgs(), &v);

	return v;
}

/*======================================================================*/
/* Get: get property via IDispatch interface							*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT Get (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	DISPPARAMS *pParams,				// parameters
	VARIANT *pRes						// result
	)
{
	return Invoke (pDisp, pName, DISPATCH_PROPERTYGET, pParams, pRes);
}

/*======================================================================*/
/* Get: get property via IDispatch interface							*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline VARIANT Get (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName					// method / property name
	)
{
	VARIANT v;
	
	HRESULT hr = Invoke (pDisp, pName, DISPATCH_PROPERTYGET, dpNoArgs(), &v);

	return v;
}

/*======================================================================*/
/* Get: get property via IDispatch interface							*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline VARIANT Get (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	DISPPARAMS *pdp						// parameters
	)
{
	VARIANT v;
	
	HRESULT hr = Invoke (pDisp, pName, DISPATCH_PROPERTYGET, pdp, &v);

	return v;
}

/*======================================================================*/
/* Put: setup property via IDispatch interface							*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT Put (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	DISPPARAMS *pParams,				// parameters
	VARIANT *pRes						// result
	)
{
	return Invoke (pDisp, pName, DISPATCH_PROPERTYPUT | DISPATCH_METHOD, pParams, pRes);
}

/*======================================================================*/
/* PutBool: setup bool property via IDispatch interface					*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT PutBool (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	BOOL b								// bool value
	)
{
	VARIANT v[1];
	v[0].vt = VT_BOOL;
	v[0].boolVal = b;					// bool value
	DISPPARAMS dp = v2dp (v);

	return Put (pDisp, pName, &dp, NULL);
}

/*======================================================================*/
/* PutBstr: setup bstr property via IDispatch interface					*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT PutBstr (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	BSTR bstr							// bstr value
	)
{
	VARIANT v[1];
	v[0].vt = VT_BSTR;
	v[0].bstrVal = bstr;				// bstr value
	DISPPARAMS dp = v2dp (v);

	return Put (pDisp, pName, &dp, NULL);
}

/*======================================================================*/
/* PutpBstr: setup pbstr property via IDispatch interface				*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT PutpBstr (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	BSTR *pbstr							// pbstr value
	)
{
	VARIANT v[1];
	v[0].vt = VT_BYREF | VT_BSTR;
	v[0].pbstrVal = pbstr;				// pbstr value
	DISPPARAMS dp = v2dp (v);

	return Put (pDisp, pName, &dp, NULL);
}

/*======================================================================*/
/* PutDouble: setup double property via IDispatch interface				*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT PutDouble (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	double d							// double value
	)
{
	VARIANT v[1];
	v[0].vt = VT_R8;
	v[0].dblVal = d;					// double value
	DISPPARAMS dp = v2dp (v);

	return Put (pDisp, pName, &dp, NULL);
}

/*======================================================================*/
/* PutFloat: setup float property via IDispatch interface				*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT PutFloat (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// method / property name
	float f								// float value
	)
{
	VARIANT v[1];
	v[0].vt = VT_R4;
	v[0].fltVal = f;					// float value
	DISPPARAMS dp = v2dp (v);

	return Put (pDisp, pName, &dp, NULL);
}

/*======================================================================*/
/* PutI4: setup integer property via IDispatch interface				*/
/* return:																*/
/*		void															*/
/*======================================================================*/

__inline HRESULT PutI4 (
	IDispatch *pDisp,					// interface
	OLECHAR FAR *pName,					// property name
	long lI4							// long value
	)
{
	VARIANT v[1];
	v[0].vt = VT_I4;
	v[0].lVal = lI4;					// long value
	DISPPARAMS dp = v2dp (v);

	return Put (pDisp, pName, &dp, NULL);
}

#endif //__Genfun_H_
