/*======================================================================*/
/* Inet.cpp: CInet Internet download object implementation				*/
/* Copyright 2001-2003, (C) Jean-Pierre Menicucci - all rights reserved	*/
/*======================================================================*/

#include "stdafx.h"								// base includes
#include "resource.h"							// resource includes
#include "wininet.h"							// wininet includes
#include "Lang.h"								// CLang resource handling class
#include "GenFun.h"								// general functions includes
#include "Down.h"								// CDown downloading dialog class
#include "Inet.h"								// CInet URL downloading class

CDown* CInet::m_pDown = 0;						// CDown downloading dialog pointer

/*======================================================================*/
/* ReadUrl : read uniform resource locator								*/
/* return:																*/
/*		0		failure													*/
/*		#0		file path												*/
/*======================================================================*/

LPSTR CInet::ReadUrl (
	LPSTR lpszUrl,								// url or file path
	BOOL* pbUrl									// is url ?
	)
{
	/********************************************/
	/* return pointer to same string if not url */
	/********************************************/

	int nUrl = lstrlen (lpszUrl);

	if (nmiCmp (lpszUrl, min (nUrl, 4), "ftp:", 4) &&
		nmiCmp (lpszUrl, min (nUrl, 5), "http:", 5) &&
		nmiCmp (lpszUrl, min (nUrl, 7), "gopher:", 7)) {
			*pbUrl = FALSE;
			return lpszUrl;
			}
	else
		*pbUrl = TRUE;

	/***********************************************/
	/* else create temporary file from url content */
	/***********************************************/

	HANDLE f = INVALID_HANDLE_VALUE;			// file handle
	HINTERNET g = 0;							// internet handle
	HINTERNET h = 0;							// internet handle
	LPSTR lp = 0;								// file name

	/*==================================================================*/
	/* use try-finally to handle resource freeing						*/
	/*==================================================================*/

	__try {

		/*****************/
		/* open internet */
		/*****************/

		h = InternetOpen ("XmlWord", INTERNET_OPEN_TYPE_PRECONFIG, 0, 0, 0);
		if (h == 0) {
			CLang::SysMesBox (IDS_ERRSYS, lpszUrl, MB_ICONSTOP | MB_OK);
			__leave;
			}

		/************/
		/* open url */
		/************/

		g = InternetOpenUrl (h, lpszUrl, 0, 0, INTERNET_FLAG_DONT_CACHE | INTERNET_FLAG_RAW_DATA, 0);
		if (g == 0) {
			CLang::SysMesBox (IDS_ERRSYS, lpszUrl, MB_ICONSTOP | MB_OK);
			__leave;
			}

		/********************************/
		/* create temporary output file */
		/********************************/

		char szTmpDir[MAX_PATH];
		if (!GetTempPath (sizeof (szTmpDir), szTmpDir)) {
			CLang::SysMesBox (IDS_ERRSYS, szTmpDir, MB_ICONSTOP | MB_OK);
			__leave;
			}

		char szTmpFil[MAX_PATH];
		if (!GetTempFileName (szTmpDir, "XXX", 0, szTmpFil)) {
			CLang::SysMesBox (IDS_ERRSYS, szTmpFil, MB_ICONSTOP | MB_OK);
			__leave;
			}

		f = CreateFile (szTmpFil, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
		if (f == INVALID_HANDLE_VALUE) {
			CLang::SysMesBox (IDS_ERRSYS, szTmpFil, MB_ICONSTOP | MB_OK);
			__leave;
			}

		/*****************************/
		/* create downloading dialog */
		/*****************************/

		CInet::m_pDown = new CDown;				// downloading dialog

		if (CInet::m_pDown)
			if (!CInet::m_pDown->Create(0)) {
				CLang::SysMesBox (GetLastError());
				__leave;
				}

		if (CInet::m_pDown && CInet::m_pDown->m_hWnd)
			SetDlgItemText (CInet::m_pDown->m_hWnd, IDC_DWNPATH, lpszUrl);

		/****************************/
		/* copy data to output file */
		/****************************/

		char szBuf[8192];						// read/write buffer
		DWORD dwBuf;							// read/write length
		DWORD dwTot;							// bytes downloaded
		dwTot = 0;

		while (TRUE) {

			/********************************/
			/* read chunk from the internet */
			/********************************/

			BOOL b = InternetReadFile (g, szBuf, sizeof (szBuf), &dwBuf);
			if (b == FALSE) {					// read error
				CLang::SysMesBox (IDS_ERRSYS, lpszUrl, MB_ICONSTOP | MB_OK);
				CloseHandle (f);
				DeleteFile (szTmpFil);
				f = INVALID_HANDLE_VALUE;
				__leave;
				}

			if (dwBuf == 0)						// end of file
				break;

			/****************************/
			/* process waiting messages */
			/****************************/

			MSG msg;
			while (::PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE)) {
				if (::GetMessage (&msg, NULL, NULL, NULL) > 0) {
					::TranslateMessage (&msg);
					::DispatchMessage (&msg);
					}
				}

			if (CInet::m_pDown && CInet::m_pDown->IsCanceled()) // user canceled
				__leave;

			/**************************************/
			/* display number of bytes downloaded */
			/**************************************/

			dwTot += dwBuf;

			char szFmt[64];
			CLang::LoadString (IDS_DWNSTAT, szFmt, sizeof (szFmt));

			char szMes[64];
			wsprintf (szMes, szFmt, dwTot);

			if (CInet::m_pDown && CInet::m_pDown->m_hWnd)
				SetDlgItemText (CInet::m_pDown->m_hWnd, IDC_DWNSIZE, szMes);

			/***********************/
			/* write chunk to file */
			/***********************/

			b = WriteFile (f, szBuf, dwBuf, &dwBuf, 0);
			if (b == FALSE) {					// write error
				CLang::SysMesBox (IDS_ERRSYS, szTmpFil, MB_ICONSTOP | MB_OK);
				CloseHandle (f);
				DeleteFile (szTmpFil);
				f = INVALID_HANDLE_VALUE;
				__leave;
				}
			}

		/****************************/
		/* close downloading dialog */
		/****************************/

		if (CInet::m_pDown) {
			if (CInet::m_pDown->m_hWnd)
				CInet::m_pDown->DestroyWindow();
			delete CInet::m_pDown;
			CInet::m_pDown = 0;
			}

		/*****************************************/
		/* store created file name in new string */
		/*****************************************/

		lp = (LPSTR) malloc (lstrlen (szTmpFil) + 1);
		lstrcpy (lp, szTmpFil);
		}

	/*==================================================================*/
	/* free resources section											*/
	/*==================================================================*/

	__finally {
		if (h) InternetCloseHandle (h);
		if (g) InternetCloseHandle (g);
		if (f != INVALID_HANDLE_VALUE) CloseHandle (f);
		}

	return lp;
}

/*======================================================================*/
/* ReadUrl : read uniform resource locator								*/
/* return:																*/
/*		0		failure													*/
/*		#0		file path												*/
/*======================================================================*/

LPSTR CInet::ReadUrl (
	LPSTR lpsUrl,								// url or file path
	int nUrl,									// length
	BOOL* pbUrl									// is url ?
	)
{
	LPSTR lpszUrl = (LPSTR) malloc (nUrl+1);	// allocate zero-terminated string
	if (lpszUrl == 0) return 0;

	nCpy (lpszUrl, lpsUrl, nUrl);				// zero-terminated string url
	*(lpszUrl+nUrl) = '\0';

	LPSTR lpszRet = ReadUrl (lpszUrl, pbUrl);	// copy to cache file if url

	free (lpszUrl);								// free string

	return lpszRet;								// return file name
}
