/*======================================================================*/
/* Lang.cpp: CLang multilanguage handling static object implementation	*/
/* Copyright 2001-2003, (C) Jean-Pierre Menicucci - all rights reserved	*/
/*======================================================================*/

#include "stdafx.h"						// base includes
#include "resource.h"					// resource includes
#include "Lang.h"						// CLang resource handling class
#include "Genfun.h"						// general functions includes

HINSTANCE CLang::m_hInst = 0;			// default instance handle
WORD CLang::m_wLangId = 0x0409;			// default language is english

/*======================================================================*/
/* FindResource: find resource using langId								*/
/* return:																*/
/*		LPWSTR							menu pointer					*/
/*======================================================================*/

HRSRC CLang::FindResource (
	LPCTSTR lpType,						// pointer to resource type 
	LPCTSTR lpName						// pointer to resource name 
	)
{
	return FindResourceEx (
		m_hInst,						// resource-module handle 
		lpType,							// pointer to resource type 
		lpName,							// pointer to resource name 
		m_wLangId						// resource language 
		);
}

/*======================================================================*/
/* GetInstance: get instance handle										*/
/* return:																*/
/*		hInst							// instance handle				*/
/*======================================================================*/

HINSTANCE CLang::GetInstance (
	void								// no argument
	)
{
	return m_hInst;						// return instance handle
}

/*======================================================================*/
/* GetLangId: get language identifier									*/
/* return:																*/
/*		wLangId							// language identifier			*/
/*======================================================================*/

WORD CLang::GetLangId (
	void								// no argument
	)
{
	return m_wLangId;					// return language identifier
}

/*======================================================================*/
/* LoadMenu: load menu using langId										*/
/* return:																*/
/*		LPWSTR							menu pointer					*/
/*======================================================================*/

LPWSTR CLang::LoadMenu (
	LPSTR lpszMenu						// menu resource number
	)
{
	HRSRC hRsrc = FindResourceEx (m_hInst, RT_MENU, lpszMenu, m_wLangId);
	HGLOBAL hDiag = hRsrc ? LoadResource (m_hInst, hRsrc) : 0;
	LPWSTR lpwStr = hDiag ? (LPWSTR) LockResource (hDiag) : 0;
	return lpwStr;
}

/*======================================================================*/
/* LoadString: LoadString using langID									*/
/* return:																*/
/*		void															*/
/*======================================================================*/

int CLang::LoadString (
	UINT uMessage,						// formatting message number
	LPTSTR lpMulti,						// receiving string
	int cchMulti,						// receiving length
	WORD wLangId						// language identifier
	)
{
	if (wLangId == 0)
		wLangId = m_wLangId;

	HRSRC hRsrc = FindResourceEx (m_hInst, RT_STRING, MAKEINTRESOURCE (1+uMessage/16), wLangId);
	HGLOBAL hDiag = LoadResource (m_hInst, hRsrc);
	LPWSTR lpwStr = (LPWSTR) LockResource (hDiag);

	if (lpwStr == NULL) {				// resource unexistent
		*lpMulti = '\0';				// return empty string
		cchMulti = 0;
		return 0;
		}

	for (WORD w = 0; w < uMessage % 16; w++)
		lpwStr += *lpwStr+1;

	int cchWide = *(lpwStr++);
	LPCWSTR lpWide = lpwStr;

	WideCharToMultiByte (
		CP_ACP,							// code page
		0,								// handling of unmapped characters
		lpWide,							// character string to be converted
		cchWide,						// number of characters in lpWide
		lpMulti,						// destination multibyte string
		cchMulti,						// number of characters in lpMulti
		NULL,							// default character
		NULL);							// default character was used

	cchMulti = min (cchWide, cchMulti-1);
	*(lpMulti+cchMulti) = '\0';

	return cchMulti;
}

/*======================================================================*/
/* MesBox: displaying a message box										*/
/* return:																*/
/*		identifier of the button pressed on quiting the message box		*/
/*======================================================================*/
/* The message is generated after a string contained in nMessage		*/
/*======================================================================*/

int CLang::MesBox (
	LPCTSTR szMsg,						// message string
	UINT fuStyle						// dialog buttons
	)
{
	TCHAR szCap[1024];
	LoadString (IDS_ERROR, szCap, sizeof (szCap));

	return MessageBox (NULL, szMsg, szCap, fuStyle);
}

/*======================================================================*/
/* MesBox: displaying a message box										*/
/* return:																*/
/*		identifier of the button pressed on quiting the message box		*/
/*======================================================================*/
/* The message is generated after a string contained in nMessage		*/
/*======================================================================*/

int CLang::MesBox (
	UINT uMsg,							// formatting message number
	UINT fuStyle						// dialog buttons
	)
{
	TCHAR szMsg[1024];
	LoadString (uMsg, szMsg, sizeof (szMsg));

	TCHAR szCap[1024];
	LoadString (IDS_ERROR, szCap, sizeof (szCap));

	return MessageBox (NULL, szMsg, szCap, fuStyle | MB_SETFOREGROUND);
}

/*======================================================================*/
/* MesBox: displaying a message box										*/
/* return:																*/
/*		identifier of the button pressed on quiting the message box		*/
/*======================================================================*/
/* The message is generated after a string contained in nMessage		*/
/*======================================================================*/

int CLang::MesBox (
	UINT uCap,							// message box caption
	UINT uMsg,							// formatting message number
	UINT fuStyle						// dialog buttons
	)
{
	TCHAR szMsg[1024];
	LoadString (uMsg, szMsg, sizeof (szMsg));

	TCHAR szCap[1024];
	LoadString (uCap, szCap, sizeof (szCap));

	return MessageBox (NULL, szMsg, szCap, fuStyle | MB_SETFOREGROUND);
}

/*======================================================================*/
/* MonthName: get month name from ascii month number					*/
/* return:																*/
/*		pointer to new allocated month name								*/
/*======================================================================*/

LPSTR CLang::MonthName (
	LPSTR szMon							// ascii month number
	)
{
	int iMon = atoi (szMon)-1;			// set up month index
	if (iMon < 0 || iMon > 11) {		// check value
		MesBox (IDS_ERROR, IDS_ERRMON, MB_ICONSTOP | MB_OK);
		return 0;
		}

	char szMons[1024];					// set up month names

	LoadString (IDS_MONTHES, szMons, sizeof (szMons));
	for (LPSTR lp = szMons; *lp; lp++) {
		if (*lp == ' ')
			*lp = '\0';
		}

	lp = szMons;
	for (int i = 0; i < iMon; i++)		// find month name
		lp += lstrlen(lp)+1;

	LPSTR lpszMon = (LPSTR) malloc (lstrlen(lp)+1);
	lstrcpy (lpszMon, lp);

	return lpszMon;
}

/*======================================================================*/
/* SetInstance: set instance handle										*/
/* return:																*/
/*		void															*/
/*======================================================================*/

void CLang::SetInstance (
	HINSTANCE hInst						// new instance handle
	)
{
	m_hInst = hInst;					// set new instance handle
}

/*======================================================================*/
/* SetLangId: set language identifier									*/
/* return:																*/
/*		void															*/
/*======================================================================*/

BOOL CALLBACK EnumResLangProc (
	HINSTANCE hModule,					// resource-module handle
	LPCTSTR lpszType,					// pointer to resource type
	LPCTSTR lpszName,					// pointer to resource name
	WORD wIDLanguage,					// resource language identifier
	LONG lParam							// application-defined parameter
	)
{
	char szLang[3];
	CLang::LoadString (IDS_LANG, szLang, sizeof (szLang), wIDLanguage);

	if (!lstrcmpi (szLang, (LPSTR) lParam)) {
		CLang::SetLangId (wIDLanguage);	// setup language
		return FALSE;					// stop enumeration
		}

	return TRUE;						// continue enumeration
}

void CLang::SetLangId (
	LPSTR lpszLang						// new language code
	)
{
	EnumResourceLanguages (
		m_hInst,						// resource-module handle
		RT_STRING,						// pointer to resource type
		MAKEINTRESOURCE(1+IDS_LANG/16),	// pointer to resource name
		EnumResLangProc,				// function to callback				
		(LONG) lpszLang					// language name
		);

	char szLang[3];
	if (!LoadString (IDS_LANG, szLang, sizeof (szLang)))
		m_wLangId = 0x0409;				// revert to english
}

/*======================================================================*/
/* SetLangId: set language identifier									*/
/* return:																*/
/*		void															*/
/*======================================================================*/

void CLang::SetLangId (
	WORD wLangId						// new language identifier
	)
{
	m_wLangId = wLangId;				// set new language identifier
}

/*======================================================================*/
/* StrMesBox: displaying a message box with a string parameter			*/
/* return:																*/
/*		identifier of the button pressed on quiting the message box		*/
/*======================================================================*/
/* The message is generated after a formatting string contained in		*/
/* nMessage and a string parameter contained in lpszString				*/
/*======================================================================*/

int CLang::StrMesBox (
	UINT uFor,							// formatting message number
	LPCTSTR lpsz,						// string parameter
	UINT fuStyle						// dialog buttons
	)
{
	TCHAR szFor[1024];					// formatting message
	LoadString (uFor, szFor, sizeof (szFor));

	TCHAR szMsg[1024];					// message buffer
	wsprintf (szMsg, szFor, lpsz);

	TCHAR szCap[1024];
	LoadString (IDS_ERROR, szCap, sizeof (szCap));

	return MessageBox (NULL, szMsg, szCap, fuStyle | MB_SETFOREGROUND);
}

/*======================================================================*/
/* StrStrMesBox: displaying a message box with two string parameters	*/
/* return:																*/
/*		identifier of the button pressed on quiting the message box		*/
/*======================================================================*/
/* The message is generated after a formatting string contained in		*/
/* nMessage and a string parameter contained in lpszString				*/
/*======================================================================*/

int CLang::StrStrMesBox (
	UINT uFor,							// formatting message number
	LPCTSTR lpsz1,						// string parameter 1
	LPCTSTR lpsz2,						// string parameter 2
	UINT fuStyle						// dialog buttons
	)
{
	TCHAR szFor[1024];					// formatting message
	LoadString (uFor, szFor, sizeof (szFor));

	TCHAR szMsg[1024];					// message buffer
	wsprintf (szMsg, szFor, lpsz1, lpsz2);

	TCHAR szCap[1024];
	LoadString (IDS_ERROR, szCap, sizeof (szCap));

	return MessageBox (NULL, szMsg, szCap, fuStyle | MB_SETFOREGROUND);
}

/*======================================================================*/
/* SysMesBox: displaying a string for GetLastError()					*/
/* return:																*/
/*		void															*/
/*======================================================================*/

void CLang::SysMesBox (
	UINT uMsg							// formatting message number
	)
{
	LPTSTR lpMsg;						// system error message

	FormatMessage (
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		NULL, uMsg, MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpMsg, 0, NULL);

	MesBox (lpMsg, MB_ICONINFORMATION | MB_OK);

	LocalFree (lpMsg);
}

/*======================================================================*/
/* SysMesBox: displaying a message box with a system error message		*/
/* return:																*/
/*		identifier of the button pressed on quiting the message box		*/
/*======================================================================*/

int CLang::SysMesBox (
	UINT uFor,							// formatting message number
	LPCTSTR lpszStr,					// string parameter
	UINT fuStyle						// dialog buttons
	)
{
	TCHAR szFor[1024];					// formatting message
	LoadString (uFor, szFor, sizeof (szFor));

	LPTSTR lpszSys;						// message buffer
	FormatMessage (
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		NULL, GetLastError(), MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpszSys, 0, NULL);

	if (lpszSys)
		*lpszSys = tolower (*lpszSys);

	TCHAR szMsg[1024];					// message buffer
	wsprintf (szMsg, szFor, lpszStr, lpszSys);

	if (lpszSys)
		LocalFree (lpszSys);

	TCHAR szCap[1024];
	LoadString (IDS_ERROR, szCap, sizeof (szCap));

	return MessageBox (NULL, szMsg, szCap, fuStyle | MB_SETFOREGROUND);
}
