/*======================================================================*/
/* Word.cpp: CWord Word automation object implementation				*/
/* Copyright 2001-2003, (C) Jean-Pierre Menicucci - all rights reserved	*/
/*======================================================================*/

#include "stdafx.h"						// base includes
#include "resource.h"					// resource includes
#include "BSTR.h"						// CBSTR BSTR wrapper class
#include "Lang.h"						// CLang resource handling class
#include "Genfun.h"						// general functions includes
#include "Word.h"						// CWord Word automation class

/*======================================================================*/
/* Add : add document from template										*/
/*======================================================================*/

HRESULT CWord::Add (
	LPSTR lpsz							// template file path
	)
{
	/********************************************************/
	/* .Documents.Add Template:=dotPath, NewTemplate:=False */
	/********************************************************/

	VARIANT d = Get (m_pDispApp, L"Documents");				// .Documents

	CBSTR bs (lpsz);
	VARIANT vDot;
	vDot.vt = VT_BSTR;
	vDot.bstrVal = bs;
	DISPPARAMS dpDot = { &vDot, 0, 1, 0 };

	Call (d.pdispVal, L"Add", &dpDot, 0);					// .Add Template:=dotPath

	/**********************/
	/* release interfaces */
	/**********************/

	d.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* CloseDocument : close document in a running word instance			*/
/*======================================================================*/

HRESULT CWord::CloseDocument (
	LPSTR lpsz							// document name
	)
{
	IBindCtx *pbc = 0;					// bind context
	IRunningObjectTable *prot = 0;		// running object table
	IEnumMoniker *pem = 0;				// enumeration interface.
	IMoniker *pmon = 0;					// entry moniker
	IDispatch *pdoc = 0;				// document
	HRESULT hr;							// return code

	/*==================================================================*/
	/* use try-finally to handle resource freeing						*/
	/*==================================================================*/

	__try {

		/***************************************/
		/* get running object table enumerator */
		/***************************************/

		hr = CreateBindCtx (0, &pbc);						// get bind context
		if (FAILED (hr)) __leave;

		hr = pbc->GetRunningObjectTable (&prot);			// get running object table
		if (FAILED (hr)) __leave;

		hr = prot->EnumRunning (&pem);						// get enumerator
		if (FAILED (hr)) __leave;

		pem->Reset();										// initialize enumeration

		/********************************/
		/* loop in running object table */
		/********************************/

		LPOLESTR pName;										// entry display name (wide char)
		char szName[MAX_PATH];								// entry display name (narrow char)

		while (pem->Next (1, &pmon, NULL) == S_OK) {

			pmon->GetDisplayName (pbc, NULL, &pName);
			WideCharToMultiByte (CP_ACP, 0, pName, -1, szName, MAX_PATH, 0, 0);

			if (!lstrcmpi (szName, lpsz)) {					// document found - bind

				/**********************************/
				/* get document IDispatch pointer */
				/**********************************/

				hr = pmon->BindToObject (pbc, NULL, IID_IDispatch, (void**) &pdoc);
				if (FAILED (hr)) __leave;

				/******************/
				/* close document */
				/******************/

				VARIANT v;
				v.vt = VT_I4;
				v.lVal = 0;									// wdDoNotSaveChanges
				DISPPARAMS dp = { &v, 0, 1, 0 };
				Call (pdoc, L"Close", &dp, 0);				// .Close

				break;
				}

			pmon->Release();
			}

		pem = 0;
		}

	/*==================================================================*/
	/* free resources section											*/
	/*==================================================================*/

	__finally {

		/**********************/
		/* release interfaces */
		/**********************/

		if (pdoc) pdoc->Release();
		if (pmon) pmon->Release();
		if (pem) pem->Release();
		if (prot) prot->Release();
		if (pbc) pbc->Release();
		}

	return hr;
}

/*======================================================================*/
/* Close : release dispatch interface									*/
/*======================================================================*/

HRESULT CWord::Close (
	void
	)
{
	/******************************/
	/* release dispatch interface */
	/******************************/

	if (m_pDispApp) {
		m_pDispApp->Release();
		m_pDispApp = 0;
		}

	return 0;
}

/*======================================================================*/
/* FieldsUpdate : update fields											*/
/*======================================================================*/

HRESULT CWord::FieldsUpdate (
	void
	)
{
	/******************************************************/
	/* .Selection.HomeKey Unit:=wdStory, .Extend=wdExtend */
	/******************************************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	VARIANT x[2];
	x[1].vt = VT_I4;										// .Unit:=wdStory
	x[1].lVal = 6;											// wdStory
	x[0].vt = VT_I4;										// .Extend:=wdExtend
	x[0].lVal = 1;											// wdExtend
	DISPPARAMS dpx = { x, 0, 2, 0 };

	Call (s.pdispVal, L"HomeKey", &dpx, 0);					// .HomeKey Unit:=wdStory, .Extend=wdExtend

	/******************/
	/* .Fields.Update */
	/******************/

	VARIANT fs = Get (s.pdispVal, L"Fields");				// .Fields

	Call (fs.pdispVal, L"Update", dpNoArgs(), 0);			// .Update

	/****************************************************/
	/* .Selection.HomeKey Unit:=wdStory, .Extend=wdMove */
	/****************************************************/

	x[0].lVal = 0;											// wdMove

	Call (s.pdispVal, L"HomeKey", &dpx, 0);					// .HomeKey Unit:=wdStory, .Extend=wdMove

	/**********************/
	/* release interfaces */
	/**********************/

	fs.pdispVal->Release();									// release interface
	s.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* GotoEnd : goto end of document										*/
/*======================================================================*/

HRESULT CWord::GotoEnd (
	void								// no argument
	)
{
	/**************************************/
	/* .Selection.EndKey Unit:=wdEndStory */
	/**************************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	VARIANT x;
	x.vt = VT_I4;											// .Unit:=wdStory
	x.lVal = 6;												// wdStory
	DISPPARAMS dpx = { &x, 0, 1, 0 };

	Call (s.pdispVal, L"EndKey", &dpx, 0);					// .EndKey Unit:=wdStory

	return 0;
}

/*======================================================================*/
/* Image : insert image													*/
/*======================================================================*/

HRESULT CWord::Image (
	LPSTR lpsz							// document name
	)
{
	/****************************/
	/* .Selection.TypeParagraph */
	/****************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	if (!m_bPageBreak)
		Call (s.pdispVal, L"TypeParagraph", dpNoArgs(), 0);	// .TypeParagraph
	else
		m_bPageBreak = FALSE;

	/**************************************/
	/* .Selection.InlineShapes.AddPicture */
	/**************************************/

	VARIANT ss = Get (s.pdispVal, L"InlineShapes");			// .InlineShapes

	VARIANT x[3];
	CBSTR bs (lpsz);
	x[2].vt = VT_BYREF | VT_BSTR;							// Filename:=lpsz
	x[2].pbstrVal = &bs;
	x[1].vt = VT_BOOL;										// LinkToFile:=false
	x[1].boolVal = FALSE;
	x[0].vt = VT_BOOL;										// SaveWithDocument=true
	x[0].boolVal = TRUE;
	DISPPARAMS dp = { x, 0, 3, 0 };

	Call (ss.pdispVal, L"AddPicture", &dp, 0);				// .AddPicture

	/*****************************************************/
	/* .ParagraphFormat.Alignment=wdAlignParagraphCenter */
	/*****************************************************/

	VARIANT pf = Get (s.pdispVal, L"ParagraphFormat");		// .ParagraphFormat

	PutI4 (pf.pdispVal, L"Alignment", 1);					// .Alignment=wdAlignParagraphCenter

	return 0;
}

/*======================================================================*/
/* LineBreak : type line break											*/
/*======================================================================*/

HRESULT CWord::LineBreak (
	void								// no argument
	)
{
	/****************************/
	/* .Selection.TypeParagraph */
	/****************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	if (!m_bPageBreak)
		Call (s.pdispVal, L"TypeParagraph", dpNoArgs(), 0);	// .TypeParagraph
	else
		m_bPageBreak = FALSE;

	/**********************/
	/* release interfaces */
	/**********************/

	s.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* Normal : type text in normal style									*/
/*======================================================================*/

HRESULT CWord::Normal (
	LPSTR lp,							// text to type
	int n								// length
	)
{
	/****************************/
	/* .Selection.TypeParagraph */
	/****************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	if (!m_bPageBreak)
		Call (s.pdispVal, L"TypeParagraph", dpNoArgs(), 0);	// .TypeParagraph
	else
		m_bPageBreak = FALSE;

	/**********************************/
	/* .Selection.Style=wdStyleNormal */
	/**********************************/

	PutI4 (s.pdispVal, L"Style", -1);						// .Style=wdStyleNormal

	/*********************************************/
	/* process string as single-styled fragments */
	/*********************************************/

	TypeFragments (s.pdispVal, lp, n);

	/**********************/
	/* release interfaces */
	/**********************/

	s.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* Open : open Word instance											*/
/*======================================================================*/

HRESULT CWord::Open (
	void
	)
{
	/****************************************/
	/* start Word and get IDispatch pointer */
	/****************************************/

	CLSID clsid;						// Word server clsid

	HRESULT hr = CLSIDFromProgID (L"Word.Application", &clsid);

	if (FAILED (hr)) {
		CLang::SysMesBox (hr);
		return hr;
		}

	IUnknown* pUnk;
	hr = GetActiveObject (clsid, NULL, (IUnknown**) &pUnk);

	if (FAILED (hr))
		hr = CoCreateInstance (clsid, 0, CLSCTX_LOCAL_SERVER, IID_IDispatch, (void **) &m_pDispApp);
	else
		hr = pUnk->QueryInterface (IID_IDispatch, (void**) &m_pDispApp);

	if (FAILED (hr)) {
		CLang::SysMesBox (hr);
		return hr;
		}

	/********************/
	/* get word version */
	/********************/

	VARIANT v = Get (m_pDispApp, L"Version");				// .Version

	int n = SysStringLen (v.bstrVal);
	LPSTR lp = (LPSTR) malloc (n);
	WideCharToMultiByte (CP_ACP, 0, v.bstrVal, n, lp, n, 0, 0);
	m_nVersion = atoi (lp);
	free (lp);

	/********************/
	/* setup word state */
	/********************/

	PutBool (m_pDispApp, L"DisplayAlerts", FALSE);			// .DisplayAlerts=false

	PutI4 (m_pDispApp, L"WindowState", 0);					// .WindowState=wdWindowStateNormal

	PutBool (m_pDispApp, L"Visible", TRUE);					// .Visible=true

	return hr;
}

/*======================================================================*/
/* PageBreak : type page break											*/
/*======================================================================*/

HRESULT CWord::PageBreak (
	void								// no argument
	)
{
	/********************************************/
	/* .Selection.InsertBreak Type:=wdPageBreak */
	/********************************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	VARIANT x;
	x.vt = VT_I4;											// Type:=wdPageBreak
	x.lVal = 7;												// wdPageBreak
	DISPPARAMS dp = { &x, 0, 1, 0 };

	Call (s.pdispVal, L"InsertBreak", &dp, 0);				// .InsertBreak Type:=wdPageBreak

	m_bPageBreak = TRUE;									// last action was a page break

	/**********************/
	/* release interfaces */
	/**********************/

	s.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* PageTitle : type text in Title-n style								*/
/*======================================================================*/

HRESULT CWord::PageTitle (
	int nStyle,							// style number
	LPSTR lp,							// text to type
	int n								// length
	)
{
	/*********************************************************************/
	/* .Selection.InsertBreak Type:=wdPageBreak (not after a page break) */
	/*********************************************************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	VARIANT x;
	DISPPARAMS dp = { &x, 0, 1, 0 };

	if (!m_bPageBreak) {

		x.vt = VT_I4;										// Type:=wdPageBreak
		x.lVal = 7;											// wdPageBreak

		Call (s.pdispVal, L"InsertBreak", &dp, 0);			// .InsertBreak Type:=wdPageBreak

		m_bPageBreak = TRUE;								// last action was a page break
		}

	/***************************************/
	/* .Selection.TypeParagraph (15 times) */
	/***************************************/

	for (int i = 0; i < 15; i++)
		Call (s.pdispVal, L"TypeParagraph", dpNoArgs(), 0);	// .TypeParagraph

	/***************************/
	/* .Selection.Style=nStyle */
	/***************************/

	PutI4 (s.pdispVal, L"Style", nStyle);					// .Style=nStyle

	/********************************/
	/* .Selection.TypeText Text:=lp */
	/********************************/

	CBSTR bsText (n, lp);
	x.vt = VT_BYREF | VT_BSTR;								// Text:=lp
	x.pbstrVal = &bsText;									// lp

	Call (s.pdispVal, L"TypeText", &dp, 0);					// .TypeText Text:=lp

	/****************************/
	/* .Selection.TypeParagraph */
	/****************************/

	Call (s.pdispVal, L"TypeParagraph", dpNoArgs(), 0);		// .TypeParagraph

	/*********************************************************************/
	/* .Selection.InsertBreak Type:=wdPageBreak (not after a page break) */
	/*********************************************************************/

	x.vt = VT_I4;											// .InsertBreak Type:=wdPageBreak
	x.lVal = 7;												// wdPageBreak

	Call (s.pdispVal, L"InsertBreak", &dp, 0);				// .InsertBreak Type:=wdPageBreak
	
	m_bPageBreak = TRUE;									// last action was a page break

	/**********************/
	/* release interfaces */
	/**********************/

	s.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* PageView : set up page view											*/
/*======================================================================*/

HRESULT CWord::PageView (
	void
	)
{
	/****************************************/
	/* .ActiveWindow.View.Type = wdPageView */
	/****************************************/

	VARIANT aw = Get (m_pDispApp, L"ActiveWindow");			// .ActiveWindow

	VARIANT v = Get (aw.pdispVal, L"View");					// .View

	PutI4 (v.pdispVal, L"Type", 3);							// .Type=wdPageView

	/**********************/
	/* release interfaces */
	/**********************/

	v.pdispVal->Release();									// release interface
	aw.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* Paste : paste object in document										*/
/*======================================================================*/

HRESULT CWord::Paste (
	float fSiz							// graph size
	)
{
	/****************************/
	/* .Selection.TypeParagraph */
	/****************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	if (!m_bPageBreak)
		Call (s.pdispVal, L"TypeParagraph", dpNoArgs(), 0);	// .TypeParagraph
	else
		m_bPageBreak = FALSE;

	/**************************************************/
	/* .Selection.PasteSpecial (paste as InlineShape) */
	/**************************************************/

	VARIANT vPsp[3];
	vPsp[0].vt = VT_I4;
	vPsp[0].lVal = 0;										// Placement:=wdInline
	vPsp[1].vt = VT_BOOL;
	vPsp[1].boolVal = 0;									// Link:=FALSE
	vPsp[2].vt = VT_ERROR;
	vPsp[2].scode = DISP_E_PARAMNOTFOUND;					// IconIndex
	DISPPARAMS dpPsp = { vPsp, 0, 3, 0 };

	Call (s.pdispVal, L"PasteSpecial", &dpPsp, 0);			// .PasteSpecial

	/*==================================================================*/
	/* Depending on word version, Object or Image is pasted as Shape	*/
	/* (floating) or InlineShape (inline). Select it to check type.		*/
	/* If InlineShape, just get it. Else convert to InlineShape.		*/
	/*==================================================================*/

	/*********************************************************************/
	/* .Selection.MoveLeft Unit:=wdCharacter, Count:=1, Extend:=wdExtend */
	/*********************************************************************/

	VARIANT vM[3];
	vM[2].vt = VT_I4;
	vM[2].lVal = 1;											// wdCharacter
	vM[1].vt = VT_I4;
	vM[1].lVal = 1;											// Count:=1
	vM[0].vt = VT_I4;
	vM[0].lVal = 1;											// Extend:=wdExtend
	DISPPARAMS dpM = { vM, 0, 3, 0 };

	Call (s.pdispVal, L"MoveLeft", &dpM);					// .Selection.MoveLeft Unit:=wdCharacter, Count:=1, Extend:=wdExtend

	/*******************/
	/* get InlineShape */
	/*******************/

	VARIANT ad = Get (m_pDispApp, L"ActiveDocument");		// .ActiveDocument

	VARIANT sp;												// Shape
	{
	VARIANT ss = Get (s.pdispVal, L"InlineShapes");			// .InlineShapes

	VARIANT c = Get (ss.pdispVal, L"Count");				// .Count

	DISPPARAMS dpc = { &c, 0, 1, 0 };

	if (c.lVal)
		Call (ss.pdispVal, L"Item", &dpc, &sp);				// .InlineShapes(.Count)

	else {

		VARIANT ss = Get (ad.pdispVal, L"Shapes");			// .Shapes

		VARIANT c = Get (ss.pdispVal, L"Count");			// .Count

		DISPPARAMS dpc = { &c, 0, 1, 0 };

		Call (ss.pdispVal, L"Item", &dpc, &sp);				// .Shapes(.Count)

		sp = Call (sp.pdispVal, L"ConvertToInlineShape");	// .ConvertToInlineShape

		ss.pdispVal->Release();								// release interface
		}

	ss.pdispVal->Release();									// release interface
	}

	/*******************************************************************/
	/* .Selection.MoveRight Unit:=wdCharacter, Count:=1 Extend:=wdMove */
	/*******************************************************************/

	vM[0].lVal = 0;											// Extend:=wdMove

	Call (s.pdispVal, L"MoveRight", &dpM);					// .Selection.MoveRight Unit:=wdCharacter, Count:=1, Extend:=wdMove

	/*****************:
	/* resize object */
	/*****************/

	fSiz /= float(100.);									// move fSiz to coefficient

	if (fSiz != float (0.)) {								// 0% - keep native size

		/******************/
		/* get page width */
		/******************/

		VARIANT ps = Get (ad.pdispVal, L"PageSetup");		// .PageSetup
		VARIANT pw = Get (ps.pdispVal, L"PageWidth");		// .PageWidth
		VARIANT lm = Get (ps.pdispVal, L"LeftMargin");		// .LeftMargin
		VARIANT rm = Get (ps.pdispVal, L"RightMargin");		// .RightMargin

		float fPagWid = pw.fltVal - lm.fltVal - rm.fltVal;	// page width

		ps.pdispVal->Release();								// release interface

		/*******************************/
		/* set object width and height */
		/*******************************/

		float fWid = fSiz * fPagWid;
		float fHei = float (.62 * fWid);

		PutFloat (sp.pdispVal, L"Width", fWid);				// .Width
		PutFloat (sp.pdispVal, L"Height", fHei);			// .Height
		}

	/*****************************************************/
	/* .ParagraphFormat.Alignment=wdAlignParagraphCenter */
	/*****************************************************/

	VARIANT pf = Get (s.pdispVal, L"ParagraphFormat");		// .ParagraphFormat

	PutI4 (pf.pdispVal, L"Alignment", 1);					// .Alignment=wdAlignParagraphCenter

	/**********************/
	/* release interfaces */
	/**********************/

	ad.pdispVal->Release();									// release interface
	pf.pdispVal->Release();									// release interface
	sp.pdispVal->Release();									// release interface
	s.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* Replace : replace all occurences of a string in a document			*/
/*======================================================================*/

HRESULT CWord::Replace (
	LPSTR lpszIn,						// string to replace
	LPSTR lpszOut						// replacing string
	)
{
	/***************************/
	/* .Selection.Find.execute */
	/***************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	VARIANT f = Get (s.pdispVal, L"Find");					// .Find

	VARIANT x[11];

	x[0].vt = VT_I4;										// Replace:=wdReplaceAll
	x[0].lVal = 2;
	CBSTR bsOut (lpszOut);
	x[1].vt = VT_BYREF | VT_BSTR;							// ReplaceWith:=lpOut
	x[1].pbstrVal = &bsOut;
	x[2].vt = VT_BOOL;										// Format:=false
	x[2].boolVal = FALSE;
	x[3].vt = VT_I4;										// Wrap:=wdFindContinue
	x[3].lVal = 1;
	x[4].vt = VT_BOOL;										// Forward:=true
	x[4].boolVal = TRUE;
	x[5].vt = VT_BOOL;										// MatchAllWordForms:=false
	x[5].boolVal = FALSE;
	x[6].vt = VT_BOOL;										// MatchSoundsLike:=false
	x[6].boolVal = FALSE;
	x[7].vt = VT_BOOL;										// MatchWildCards:=false
	x[7].boolVal = FALSE;
	x[8].vt = VT_BOOL;										// MatchWholeWord:=false
	x[8].boolVal = FALSE;
	x[9].vt = VT_BOOL;										// MatchCase:=false
	x[9].boolVal = FALSE;
	x[10].vt = VT_BSTR;										// FindText:=lpszIn
	x[10].bstrVal = CBSTR (lpszIn);

	DISPPARAMS dp = { x, 0, 11, 0 };

	Call (f.pdispVal, L"Execute", &dp, 0);					// .Execute

	/**********************/
	/* release interfaces */
	/**********************/

	f.pdispVal->Release();									// release interface
	s.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* Save : save file														*/
/*======================================================================*/

HRESULT CWord::Save (
	void
	)
{
	/************************/
	/* .ActiveDocument.Save */
	/************************/

	VARIANT ad = Get (m_pDispApp, L"ActiveDocument");		// .ActiveWindow

	Call (ad.pdispVal, L"Save", dpNoArgs(), 0);				// .Save

	/**********************/
	/* release interfaces */
	/**********************/

	ad.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* SaveAs : save file as												*/
/*======================================================================*/

HRESULT CWord::SaveAs (
	LPSTR lpszPath						// file path
	)
{
	/**************************/
	/* .ActiveDocument.SaveAs */
	/**************************/

	VARIANT x;

	CBSTR bs (lpszPath);
	x.vt = VT_BYREF | VT_BSTR;								// .FileName:=bs
	x.pbstrVal = &bs;

	DISPPARAMS dp = { &x, 0, 1, 0 };

	VARIANT ad = Get (m_pDispApp, L"ActiveDocument");		// .ActiveWindow

	Call (ad.pdispVal, L"SaveAs", &dp, 0);					// .SaveAs

	/**********************/
	/* release interfaces */
	/**********************/

	ad.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* SeekView : activate view												*/
/*======================================================================*/

HRESULT CWord::SeekView (
	int nView							// view to activate
	)
{
	/**************************************************/
	/* .ActiveWindow.ActivePane.View.SeekView = nView */
	/**************************************************/

	VARIANT aw = Get (m_pDispApp, L"ActiveWindow");			// .ActiveWindow
	VARIANT ap = Get (aw.pdispVal, L"ActivePane");			// .ActivePane
	VARIANT v = Get (ap.pdispVal, L"View");					// .View

	PutI4 (v.pdispVal, L"SeekView", nView);					// .SeekView=nView

	/**********************/
	/* release interfaces */
	/**********************/

	v.pdispVal->Release();									// release interface
	ap.pdispVal->Release();									// release interface
	aw.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* Table : type table													*/
/*======================================================================*/

HRESULT CWord::Table (
	LPSTR lpsSiz, int nSizLen,			// Column widthes
	LPSTR lpsWid, int nWidLen,			// Column widthes
	LPSTR lpsAli, int nAliLen,			// Column alignments
	LPSTR lpsDat, int nDatLen			// Table data
	)
{
	/******************************************/
	/* .ActiveWindow.View.Type = wdNormalView */
	/******************************************/

	VARIANT aw = Get (m_pDispApp, L"ActiveWindow");			// .ActiveWindow

	VARIANT v = Get (aw.pdispVal, L"View");					// .View

	PutI4 (v.pdispVal, L"Type", 1);							// .Type=wdNormalView

	/***********/
	/* no data */
	/***********/

	if (nLfs (lpsDat, nDatLen) < 2) {
		char szMsg[MAX_PATH];
		int	nMsg = CLang::LoadString (IDS_NOTABLE, szMsg, sizeof (szMsg));
		Normal (szMsg, nMsg);
		return 0;
		}

	/****************/
	/* column count */
	/****************/

	int nCol = nSpaces (lpsWid, nWidLen) + 1;				// column count

	/****************************/
	/* .Selection.TypeParagraph */
	/****************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	if (!m_bPageBreak)
		Call (s.pdispVal, L"TypeParagraph", dpNoArgs(), 0);	// .TypeParagraph
	else
		m_bPageBreak = FALSE;

	/**************************************************/
	/* .Selection.TypeParagraph.Style = wdStyleNormal */
	/**************************************************/

	PutI4 (s.pdispVal, L"Style", -1);						// .Style=wdStyleNormal

	/*****************************************/
	/* .Selection.Font.Size = .Font.Size - 1 */
	/*****************************************/

	VARIANT f = Get (s.pdispVal, L"Font");					// .Font

	VARIANT z = Get (f.pdispVal, L"Size");					// .Size

	PutFloat (f.pdispVal, L"Size", z.fltVal-float(1.));		// =.Font.Size-1

	/*********************************************************************/
	/* .Selection.Tables.Add Range:=.Range, NumRows:=1, NumColumns:=nCol */
	/*********************************************************************/

	VARIANT rg = Get (s.pdispVal, L"Range");				// .Range

	VARIANT x[10];
	x[2] = rg;												// Range:=.Range
	x[1].vt = VT_I4;
	x[1].lVal = 1;											// NumRows:=1
	x[0].vt = VT_I4;
	x[0].lVal = nCol;										// NumCols:=nCol
	DISPPARAMS dp = { x, 0, 3, 0 };

	VARIANT ts = Get (s.pdispVal, L"Tables");				// .Tables

	Call (ts.pdispVal, L"Add", &dp, 0);						// .Add

	/******************/
	/* type data rows */
	/******************/

	x[0].vt = VT_I4;
	x[0].lVal = 12;											// Unit:=wdCell
	dp.cArgs = 1;

	for (int iRow = 0; nDatLen > 0; iRow++) {

		for (int nDat = 0; nDat < nDatLen; nDat++)			// get line length
			if (*(lpsDat+nDat) == '\x0A')
				break;

		LPSTR lpsCel = lpsDat;
		int nCelLen = nDat;

		for (int iCol = 0; iCol < nCol; iCol++) {

			/*************************************/
			/* .Selection.MoveRight Unit:=wdCell */
			/*************************************/

			if (iRow || iCol)
				Call (s.pdispVal, L"MoveRight", &dp, 0);	// .Move Unit:=wdCell, Count:=1

			for (int nCel = 0; nCel < nCelLen; nCel++)		// get cell length
				if (*(lpsCel+nCel) == '\x09')
					break;

			TypeCell (s.pdispVal, lpsCel, nCel);			// TypeCell

			lpsCel += nCel+1;								// move to next cell
			nCelLen -= nCel+1;
			}

		lpsDat += nDat+1;									// move to next line
		nDatLen -= nDat+1;
		}

	/*******************************/
	/* .Selection.Tables(1).Select */
	/*******************************/

	x[0].vt = VT_I4;
	x[0].lVal = 1;
	dp.cArgs = 1;

	VARIANT t = Call (ts.pdispVal, L"Item", &dp);			// .Tables(1)

	Call (t.pdispVal, L"Select", dpNoArgs(), 0);			// .Select

	/***********************************************/
	/* .Selection.Rows.AllowBreakAcrossPages=false */
	/***********************************************/
	{
	VARIANT rs = Get (s.pdispVal, L"Rows");					// . Rows

//	PutI4 (rs.pdispVal, L"AllowBreakAcrossPages", 0);		// .AllowBreakAcrossPages=false

	rs.pdispVal->Release();									// release interface
	}
	/**************************************************************/
	/* .Selection.Tables(1).AutoFormat Format:=wdTableFormatGrid5 */
	/**************************************************************/

	x[9].vt = VT_I4;
	x[9].lVal = 20;											// Format:=wdTableFormatGrid5
//	x[9].lVal = 35;											// Format:=wdTableFormatContemporary
	x[8].vt = VT_BOOL;
	x[8].boolVal = TRUE;									// ApplyBorders:=true
	x[7].vt = VT_BOOL;
	x[7].boolVal = TRUE;									// ApplyShading:=true
	x[6].vt = VT_BOOL;
	x[6].boolVal = TRUE;									// ApplyFont:=true
	x[5].vt = VT_BOOL;
	x[5].boolVal = TRUE;									// ApplyColor:=true
	x[4].vt = VT_BOOL;
	x[4].boolVal = TRUE;									// ApplyHeadingRows:=true
	x[3].vt = VT_BOOL;
	x[3].boolVal = FALSE;									// ApplyLastRow:=false
	x[2].vt = VT_BOOL;
	x[2].boolVal = FALSE;									// ApplyFirstColumn:=false
	x[1].vt = VT_BOOL;
	x[1].boolVal = FALSE;									// ApplyLastColumn:=false
	x[0].vt = VT_BOOL;
	x[0].boolVal = FALSE;									// AutoFit:=false
	dp.cArgs = 10;

	Call (t.pdispVal, L"Autoformat", &dp, 0);				// .Autoformat Format=wdTableFormatGrid5

	/****************************************************************/
	/* .Selection.Cells.VerticalAlignment=wdCellAlignVerticalCenter */
	/****************************************************************/

	VARIANT cs = Get (s.pdispVal, L"Cells");				// .Cells

	PutI4 (cs.pdispVal, L"VerticalAlignment", 1);			// .VerticalAlignment=wdCellAlignVerticalCenter

	/**********************************************/
	/* .Selection. .ParagraphFormat.SpaceAfter=0. */
	/**********************************************/

	VARIANT pf = Get (s.pdispVal, L"ParagraphFormat");		// .ParagraphFormat
	
	PutFloat (pf.pdispVal, L"SpaceAfter", float(0.));		// .ParagraphFormat.SpaceAfter=0.

	/******************/
	/* get page width */
	/******************/

	VARIANT ad = Get (m_pDispApp, L"ActiveDocument");		// .ActiveDocument
	VARIANT ps = Get (ad.pdispVal, L"PageSetup");			// .PageSetup
	VARIANT pw = Get (ps.pdispVal, L"PageWidth");			// .PageWidth
	VARIANT lm = Get (ps.pdispVal, L"LeftMargin");			// .LeftMargin
	VARIANT rm = Get (ps.pdispVal, L"RightMargin");			// .RightMargin

	float fPagWid = pw.fltVal - lm.fltVal - rm.fltVal;		// page width

	/*******************************************************/
	/* set column widthes, alignment table, column headers */
	/*******************************************************/

	x[1].vt = VT_R4;
	x[1].fltVal = float (0.);								// ColumnWidth:=fColWid
	x[0].vt = VT_I4;
	x[0].lVal = 0;											// RulerStyle:=wdAdjustNone
	dp.cArgs = 2;

	VARIANT y;
	y.vt = VT_I4;
	DISPPARAMS dq = { &y, 0, 1, 0 };
	
	for (int iCol = 0; iCol < nCol; iCol++) {

		/*********************************************/
		/* .Selection.Tables(1).Columns(iCol).Select */
		/*********************************************/

		VARIANT os = Get (t.pdispVal, L"Columns");			// .Columns

		y.lVal = iCol+1;									// Index:=iCol

		VARIANT o = Call (os.pdispVal, L"Item", &dq);		// .Columns(iCol)

		Call (o.pdispVal, L"Select", dpNoArgs(), 0);		// .Select

		/****************/
		/* column width */
		/****************/

		x[1].fltVal = (float) (((atof (lpsWid)) * fPagWid + 50.) / 100.);

		Call (cs.pdispVal, L"SetWidth", &dp, 0);			// .SetWidth ColumnWidth:=fColWid RulerStyle:=wdAdjustNone

		while (*lpsWid != ' ' && *lpsWid != '<')
			lpsWid++;
		lpsWid++;

		/********************/
		/* column alignment */
		/********************/

		for (int nAli = 0; nAli < nAliLen; nAli++)			// get field length
			if (*(lpsAli+nAli) == ' ')
				break;

		if (!nmiCmp (lpsAli, nAli, "left", 4))
			PutI4 (pf.pdispVal, L"Alignment", 0);			// .ParagraphFormat.Alignment=wdAlignParagraphLeft
		else if (!nmiCmp (lpsAli, nAli, "center", 6))
			PutI4 (pf.pdispVal, L"Alignment", 1);			// .ParagraphFormat.Alignment=wdAlignParagraphCenter
		else
			PutI4 (pf.pdispVal, L"Alignment", 2);			// .ParagraphFormat.Alignment=wdAlignParagraphRight

		o.pdispVal->Release();								// release interface

		os.pdispVal->Release();								// release interface

		lpsAli += nAli+1;									// move to next field
		nAliLen -= nAli+1;
		}

	/***************************************/
	/* .Selection.Tables(1).Rows(1).Select */
	/***************************************/

	VARIANT rs = Get (t.pdispVal, L"Rows");					// . Rows

	x[0].vt = VT_I4;
	x[0].lVal = 1;
	dp.cArgs = 1;

	VARIANT r = Call (rs.pdispVal, L"Item", &dp);			// .Rows(1)

	Call (r.pdispVal, L"Select", dpNoArgs(), 0);			// .Select

	/***************************************************/
	/* .Selection.Tables(1).Rows(1).HeadingFormat=true */
	/***************************************************/

	if (m_nVersion == 8)									// works only with word 97
		PutI4 (r.pdispVal, L"HeadingFormat", 1);			// .HeadingFormat=true

	/***************************************************************/
	/* .Selection.ParagraphFormat.Alignment=wdAlignParagraphCenter */
	/***************************************************************/

	PutI4 (pf.pdispVal, L"Alignment", 1);					// .ParagraphFormat.Alignment=wdAlignParagraphCenter

	/*****************************/
	/* .Selection.Font.Bold=true */
	/*****************************/

	PutBool (f.pdispVal, L"Bold", TRUE);					// .Selection.Font.Bold:=true

	/*******************************/
	/* .Selection.Tables(1).Select */
	/*******************************/

	Call (t.pdispVal, L"Select", dpNoArgs(), 0);			// .Select

	/******************************************/
	/* .MoveRight Unit:=wdCharacter, Count:=1 */
	/******************************************/

	x[1].vt = VT_I4;
	x[1].lVal = 1;											// Unit:=wdCharacter
	x[0].vt = VT_I4;
	x[0].lVal = 1;											// Count:=1
	dp.cArgs = 2;

	Call (s.pdispVal, L"MoveRight", &dp, 0);				// .MoveRight Unit:=wdCharacter, Count:=1

	/**********************/
	/* release interfaces */
	/**********************/

	r.pdispVal->Release();									// release interface
	rs.pdispVal->Release();									// release interface
	ps.pdispVal->Release();									// release interface
	ad.pdispVal->Release();									// release interface
	pf.pdispVal->Release();									// release interface
	cs.pdispVal->Release();									// release interface
	t.pdispVal->Release();									// release interface
	ts.pdispVal->Release();									// release interface
	rg.pdispVal->Release();									// release interface
	f.pdispVal->Release();									// release interface
	s.pdispVal->Release();									// release interface
	v.pdispVal->Release();									// release interface
	aw.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* Title : type text in Title-n style									*/
/*======================================================================*/

HRESULT CWord::Title (
	int nStyle,							// style number
	LPSTR lp,							// text to type
	int n								// length
	)
{
	/****************************/
	/* .Selection.TypeParagraph */
	/****************************/

	VARIANT s = Get (m_pDispApp, L"Selection");				// .Selection

	if (!m_bPageBreak)
		Call (s.pdispVal, L"TypeParagraph", dpNoArgs(), 0);	// .TypeParagraph
	else
		m_bPageBreak = FALSE;

	/***************************/
	/* .Selection.Style=nStyle */
	/***************************/

	PutI4 (s.pdispVal, L"Style", nStyle);					// .Style=nStyle

	/********************************/
	/* .Selection.TypeText Text:=lp */
	/********************************/

	VARIANT x;
	CBSTR bsText (n, lp);
	x.vt = VT_BYREF | VT_BSTR;								// Text:=lp
	x.pbstrVal = &bsText;									// lp
	DISPPARAMS dp = { &x, 0, 1, 0 };

	Call (s.pdispVal, L"TypeText", &dp, 0);					// .TypeText Text:=lp

	/**********************/
	/* release interfaces */
	/**********************/

	s.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* TypeCell : type text in table cell									*/
/*======================================================================*/

HRESULT CWord::TypeCell (
	IDispatch *ps,						// selection
	LPSTR lp,							// text to type
	int n								// length
	)
{
	/*********************************/
	/* type \r\n delimited fragments */
	/*********************************/

	VARIANT x;
	x.vt = VT_BYREF | VT_BSTR;
	DISPPARAMS dp = v2dp (&x);

	for (int iFirst = 0, iLast = 0; iLast < n;) {

		if (!nmiCmp (lp+iLast, min (2, n-iLast), "\\r", 2) ||
			!nmiCmp (lp+iLast, min (2, n-iLast), "\\n", 2)) {

			if (iLast > iFirst) {

				TypeFragments (ps, lp+iFirst, iLast-iFirst);

				Call (ps, L"TypeParagraph", dpNoArgs(),0);	// .TypeParagraph
				}

			iLast += 2;					// skip \r or \n
			iFirst = iLast;
			}

		else
			iLast++;
		}

	/***********************/
	/* write last fragment */
	/***********************/

	if (iLast > iFirst)
		TypeFragments (ps, lp+iFirst, iLast-iFirst);

	return 0;
}

/*======================================================================*/
/* TypeFragments : type fragments in different formats					*/
/*======================================================================*/

HRESULT CWord::TypeFragments (
	IDispatch* ps,						// selection
	LPSTR lp,							// string
	int n								// length
	)
{
	/*********************************************/
	/* process string as single-styled fragments */
	/*********************************************/

	VARIANT f = Get (ps, L"Font");							// .Font

	VARIANT x;
	x.vt = VT_BYREF | VT_BSTR;
	DISPPARAMS dp = v2dp (&x);
	int j;

	for (int i = 0; i < n; i = j) {

		if (*(lp+i) == '<') {								// tag fragment
			for (j = i+1; j < n; j++) {
				if (*(lp+j) == '>') {						// end of tag fragment
					j++;
					break;
					}
				}

			if (!nmiCmp (lp+i, j-i, "<b>", 3))
				PutBool (f.pdispVal, L"Bold", TRUE);		// .Bold:=true
			else if (!nmiCmp (lp+i, j-i, "</b>", 4))
				PutBool (f.pdispVal, L"Bold", FALSE);		// .Bold:=false
			else if (!nmiCmp (lp+i, j-i, "<u>", 3))
				PutI4 (f.pdispVal, L"Underline", 1);		// .Underline:=wdUnderlineSingle
			else if (!nmiCmp (lp+i, j-i, "</u>", 4))
				PutI4 (f.pdispVal, L"Underline", 0);		// .Underline:=wdUnderlineNone
			else if (!nmiCmp (lp+i, j-i, "<i>", 3))
				PutBool (f.pdispVal, L"Italic", TRUE);		// .Italic:=true
			else if (!nmiCmp (lp+i, j-i, "</i>", 4))
				PutBool (f.pdispVal, L"Italic", FALSE);		// .Italic:=false
			}

		else {												// non-tag fragment
			for (j = i+1; j < n; j++) {
				if (*(lp+j) == '<') {						// end of tag fragment
					break;
					}
				}

			CBSTR bsText (j-i, lp+i);
			x.pbstrVal = &bsText;
			Put (ps, L"TypeText", &dp, 0);					// .TypeText Text:=bsText
			}
		}

	/**********************/
	/* release interfaces */
	/**********************/

	f.pdispVal->Release();									// release interface

	return 0;
}

/*======================================================================*/
/* Zoom : set up page zoom percentage									*/
/*======================================================================*/

HRESULT CWord::Zoom (
	int nZoom							// percentage
	)
{
	/*********************************************************/
	/* .ActiveWindow.ActivePane.View.Zoom.Percentage = nZoom */
	/*********************************************************/

	VARIANT aw = Get (m_pDispApp, L"ActiveWindow");			// .ActiveWindow
	VARIANT ap = Get (aw.pdispVal, L"ActivePane");			// .ActivePane
	VARIANT v = Get (ap.pdispVal, L"View");					// .View
	VARIANT z = Get (v.pdispVal, L"Zoom");					// .Zoom

	PutI4 (z.pdispVal, L"Percentage", nZoom);				// .Percentage=nZoom

	/**********************/
	/* release interfaces */
	/**********************/

	z.pdispVal->Release();									// release interface
	v.pdispVal->Release();									// release interface
	ap.pdispVal->Release();									// release interface
	aw.pdispVal->Release();									// release interface

	return 0;
}
