/*======================================================================*/
/* Xml.cpp: CXml parsing object implementation							*/
/* Copyright 2001-2003, (C) Jean-Pierre Menicucci - all rights reserved	*/
/*======================================================================*/

#include "stdafx.h"						// base includes
#include "resource.h"					// resource includes
#include "Lang.h"						// CLang resource handling class
#include "Genfun.h"						// general functions includes
#include "Xml.h"						// CXml XML parsing class

/*======================================================================*/
/* Accents : transcode XML accented letters to ANSI						*/
/*======================================================================*/

void CXml::Accents (
	void
	)
{
	int i = 0;							// input index
	int o = 0;							// output index

	while (i < m_nLen) {

		/**********/
		/* entity */
		/**********/

		LPSTR lpszi = m_lpsz+i;
		LPSTR lpszo = m_lpsz+o;
		int ni = m_nLen-i;

		if (*lpszi == '&') {

			if (!nmCmp (lpszi, min (8, ni), "&eacute;", 8))
				{ *(m_lpsz+o) = ''; i += 8; o++; continue; }
			else if (!nmCmp (lpszi, min (8, ni), "&egrave;", 8))
				{ *(m_lpsz+o) = ''; i += 8; o++; continue; }
			else if (!nmCmp (lpszi, min (8, ni), "&ccedil;", 8))
				{ *(m_lpsz+o) = ''; i += 8; o++; continue; }
			else if (!nmCmp (lpszi, min (8, ni), "&agrave;", 8))
				{ *(m_lpsz+o) = ''; i += 8; o++; continue; }
			else if (!nmCmp (lpszi, min (8, ni), "&ugrave;", 8))
				{ *(m_lpsz+o) = ''; i += 8; o++; continue; }
			else if (!nmCmp (lpszi, min (6, ni), "&uuml;", 6))
				{ *(m_lpsz+o) = ''; i += 6; o++; continue; }
			else if (!nmCmp (lpszi, min (7, ni), "&ocirc;", 7))
				{ *(m_lpsz+o) = ''; i += 7; o++; continue; }
			else {
				*(m_lpsz+o) = *(lpszi);
				i++;
				o++;
				}
			}

		/********/
		/* crlf */
		/********/

		else if (*lpszi == '\r') {
			if (!nmCmp (lpszi, min (2, ni), "\r\n", 2))
				{ *(m_lpsz+o) = '\n'; i += 2; o++; continue; }
			else {
				*(m_lpsz+o) = *(lpszi);
				i++;
				o++;
				}
			}

		/*********/
		/* other */
		/*********/

		else {
			*(m_lpsz+o) = *(lpszi);
			i++;
			o++;
			}
		}

	m_nLen = o;
}

/*======================================================================*/
/* Entities : transcode XML entities to ANSI characters					*/
/*======================================================================*/

void CXml::Entities (
	void
	)
{
	int i = 0;							// input index
	int o = 0;							// output index

	while (i < m_nLen) {

		/**********/
		/* entity */
		/**********/

		LPSTR lpszi = m_lpsz+i;
		LPSTR lpszo = m_lpsz+o;
		int ni = m_nLen-i;

		if (*lpszi == '&') {

			if (!nmCmp (lpszi, min (4, ni), "&lt;", 4))
				{ *(m_lpsz+o) = '<'; i += 4; o++; continue; }
			else if (!nmCmp (lpszi, min (4, ni), "&gt;", 4))
				{ *(m_lpsz+o) = '>'; i += 4; o++; continue; }
			else if (!nmCmp (lpszi, min (5, ni), "&amp;", 5))
				{ *(m_lpsz+o) = '&'; i += 5; o++; continue; }
			else if (!nmCmp (lpszi, min (6, ni), "&apos;", 6))
				{ *(m_lpsz+o) = '\''; i += 6; o++; continue; }
			else if (!nmCmp (lpszi, min (6, ni), "&quot;", 6))
				{ *(m_lpsz+o) = '"'; i += 6; o++; continue; }
			}

		/*********/
		/* other */
		/*********/

		*(m_lpsz+o) = *(lpszi);
		i++;
		o++;
		}

	m_nLen = o;
}

/*======================================================================*/
/* Parse : get next high level tag and included data					*/
/* return :																*/
/*		0		success													*/
/*		-1		no more data											*/
/*======================================================================*/

int CXml::Parse (
	int* pnTagOff,						// tag offset
	int* pnTagLen,						// tag length
	int* pnDatOff,						// data offset
	int* pnDatLen,						// data length
	int* pnAttOff,						// attribute offset
	int* pnAttLen						// attribute length
	)
{
	/******************************************/
	/* get tag, skipping dtd and comment tags */
	/******************************************/

	while (TRUE) {

		/*******************************/
		/* find tag begining delimiter */
		/*******************************/

		for (; m_nOff < m_nLen; m_nOff++) {
			if (*(m_lpsz+m_nOff) == '<') {
				m_nOff++;
				break;
				}
			}

		if (m_nOff >= m_nLen) {			// end of string - no tag
			*pnTagOff = 0;
			*pnTagLen = 0;
			*pnDatOff = 0;
			*pnDatLen = 0;
			if (pnAttOff)
				*pnAttOff = 0;
			if (pnAttLen)
				*pnAttLen = 0;
			return -1;
			}

		*pnTagOff = m_nOff;				// tag offset

		/*****************************/
		/* find tag ending delimiter */
		/*****************************/

		for (; m_nOff < m_nLen; m_nOff++) {
			if (*(m_lpsz+m_nOff) == '>') {
				m_nOff++;
				break;
				}
			}

		if (m_nOff >= m_nLen) {			// end of string - no ending delimiter
			char szExt[32];				// xml extract
			nmCpy (szExt, sizeof (szExt), m_lpsz+*pnTagOff-1, m_nLen-*pnTagOff+1);
			CLang::StrMesBox (IDS_ERRENDDEL, szExt);
			return -1;
			}

		*pnTagLen = m_nOff-*pnTagOff-1;	// tag length

		if (*(m_lpsz+*pnTagOff) == '?')	// dtd tag - skip
			continue;

		if (*(m_lpsz+*pnTagOff) == '!')	// comment - skip
			continue;

		break;
		}

	/******************************/
	/* process self-contained tag */
	/******************************/

	if (*(m_lpsz+*pnTagLen+*pnTagOff-1) == '/') {
		(*pnTagLen)--;
		*pnDatOff = 0;
		*pnDatLen = 0;
		return 0;
		}

	/***********************/
	/* set up data pointer */
	/***********************/

	*pnDatOff = *pnTagOff+*pnTagLen+1;

	/**********************************/
	/* process first attribute if any */
	/**********************************/

	if (pnAttOff && pnAttLen) {

		*pnAttOff = 0;
		*pnAttLen = 0;

		for (int nTagLen = 0; nTagLen < *pnTagLen; nTagLen++) {

			if (*(m_lpsz+*pnTagOff+nTagLen) == ' ') {

				for (int nAttOff = nTagLen+1; nAttOff < *pnTagLen; nAttOff++) {

					if (*(m_lpsz+*pnTagOff+nAttOff) == '"') {

						for (int nAttLen = nAttOff+1; nAttLen < *pnTagLen; nAttLen++) {

							if (*(m_lpsz+*pnTagOff+nAttLen) == '"') {

								*pnAttLen = nAttLen - nAttOff - 1;
								*pnAttOff = *pnTagOff + nAttOff + 1;
								*pnTagLen = nTagLen;

								break;
								}
							}
						break;
						}
					}
				break;
				}
			}
		}

	/*******************/
	/* find ending tag */
	/*******************/

	int nOff;							// worker offset
	int nLen;							// worker length

	while (TRUE) {

		/*******************************/
		/* find tag begining delimiter */
		/*******************************/

		for (; m_nOff < m_nLen; m_nOff++) {
			if (*(m_lpsz+m_nOff) == '<') {
				m_nOff++;
				break;
				}
			}

		if (m_nOff >= m_nLen) {			// end of string - ending tag not found
			char szTag[32];				// xml string extract
			nmCpy (szTag, sizeof (szTag), m_lpsz+*pnTagOff, *pnTagLen);
			CLang::StrStrMesBox (IDS_ERRENDTAG, szTag, szTag);
			return -1;
			}

		nOff = m_nOff;					// tag offset

		/*****************************/
		/* find tag ending delimiter */
		/*****************************/

		for (; m_nOff < m_nLen; m_nOff++) {
			if (*(m_lpsz+m_nOff) == '>') {
				m_nOff++;
				break;
				}
			}

		if (m_nOff >= m_nLen &&
		*(m_lpsz+m_nOff-1) != '>') {	// end of string - no ending delimiter
			char szXml[32];				// xml string extract
			nmCpy (szXml, sizeof (szXml), m_lpsz+nOff-1, m_nLen-nOff+1);
			CLang::StrMesBox (IDS_ERRENDDEL, szXml);
			return -1;
			}

		nLen = m_nOff-nOff-1;			// tag length

		if (*(m_lpsz+nOff) != '/')		// no slash - not an ending tag
			continue;

		if (m_nOff-nOff < 3)			// too short - not an ending tag
			continue;

		if (nmCmp (m_lpsz+*pnTagOff, *pnTagLen, m_lpsz+nOff+1, nLen-1))
			continue;

		break;
		}

	/**********************/
	/* set up data length */
	/**********************/

	*pnDatLen = nOff-*pnDatOff-1;

	/*****************************/
	/* strip leading crs and lfs */
	/*****************************/

	for (int nCrsOrLfs = 0; nCrsOrLfs < *pnDatLen; nCrsOrLfs++) {
		if (*(m_lpsz+*pnDatOff+nCrsOrLfs) != '\x0A' &&
			*(m_lpsz+*pnDatOff+nCrsOrLfs) != '\x0D')
			break;
		}

	*pnDatOff += nCrsOrLfs;
	*pnDatLen -= nCrsOrLfs;

	return 0;
}
