/*======================================================================*/
/* XmlWordAx.cpp : CXmlWordAx ActiveX implementation					*/
/* Copyright 2001-2003, (C) Jean-Pierre Menicucci - all rights reserved	*/
/*======================================================================*/

#include "stdafx.h"								// base includes
#include "resource.h"							// resource includes
#include "Lang.h"								// CLang resource handling class
#include "Genfun.h"								// general functions includes
#include "Down.h"								// CDown downloading dialog class
#include "Inet.h"								// CInet URL downloading class
#include "Xml.h"								// CXml XML parsing class
#include "Excel.h"								// CExcel Excel automation class
#include "Word.h"								// CWord Word automation class
#include "XmlWord.h"							// automation server
#include "XmlWordAx.h"							// XmlWord ActiveX object
#include <process.h>							// _beginthread

/////////////////////////////////////////////////////////////////////////////
// CXmlWordAx

/*======================================================================*/
/* OnDraw : ATL generated default OnDraw method (not used here)			*/
/*======================================================================*/

HRESULT CXmlWordAx::OnDraw (ATL_DRAWINFO& di)
{
	RECT& rc = *(RECT*) di.prcBounds;
	Rectangle (di.hdcDraw, rc.left, rc.top, rc.right, rc.bottom);
	DrawText (di.hdcDraw, _T("XmlWord"), -1, &rc, DT_CENTER | DT_VCENTER | DT_SINGLELINE);
	return S_OK;
}

/*======================================================================*/
/* JustDoIt : method to launch processing in the background				*/
/*======================================================================*/

void JustDoIt (void* pvParams);					// forward declaration

STDMETHODIMP CXmlWordAx::JustDoIt (BSTR bsXmlFile)
{
	// TODO: Add your implementation code here

	/*************************************/
	/* store strings in single parameter */
	/*************************************/

	int nXmlFile = SysStringLen (bsXmlFile);

	LPSTR lpszParams = (LPSTR) malloc (nXmlFile + 1);
	if (lpszParams == 0) {
		CLang::MesBox (IDS_ERRALLOC);
		return S_FALSE;
		}

	WideCharToMultiByte (CP_ACP, 0, bsXmlFile, nXmlFile, lpszParams, nXmlFile, 0, 0);
	*(lpszParams + nXmlFile) = '\0';

	/**************************/
	/* launch separate thread */
	/**************************/

	_beginthread (::JustDoIt, 0, (void*) lpszParams);

	return S_OK;
}

/*======================================================================*/
/* JustDoIt : method to effect processing in the background				*/
/*======================================================================*/

void JustDoIt (void* pvParams)
{
#ifdef DEBUG
	_CrtMemState state;
	_CrtMemCheckpoint (&state);					// check out memory leaks
#endif

	/******************************************************/
	/* variable dynamically allocated and __finally freed */
	/******************************************************/

	LPSTR lpXml = 0;							// xml buffer pointer

	BOOL bOleInitialized = FALSE;				// Ole initialized

	LPSTR lpSub = 0;							// substitution buffer

	/*==================================================================*/
	/* using try-finally to manage resources freeing					*/
	/*==================================================================*/

	__try {

		/*****************************/
		/* read xml from url or file */
		/*****************************/

		DWORD dwXml;							// xml buffer length
		{
		LPSTR lpszPath = (LPSTR) pvParams;		// xml file path
		BOOL bPathIsUrl;						// path is url

		if ((lpszPath = CInet::ReadUrl (lpszPath, &bPathIsUrl)) == 0)
			return;

		int nError = FileRead (lpszPath, &lpXml, &dwXml);

		if (bPathIsUrl) {
			DeleteFile (lpszPath);				// delete temporary file
			free (lpszPath);					// free allocated path
			}

		free (pvParams);						// not required after this point

		if (nError)								// file read error
			return;
		}
		/*************/
		/* parse xml */
		/*************/

		CXml Xml (lpXml, (int) dwXml);			// set up xml object

		Xml.Accents();							// change xml accented letters to ANSI

		/*************************/
		/* get dashboard element */
		/*************************/

		int nTagOff, nTagLen;					// tag
		int nDatOff, nDatLen;					// data
		int nAttOff, nAttLen;					// attribute

		if (Xml.Parse (&nTagOff, &nTagLen, &nDatOff, &nDatLen, &nAttOff, &nAttLen))
			return;

		LPSTR lpDash = lpXml + nDatOff;
		int nDashLen = nDatLen;
		CXml Dash (lpDash, nDashLen);			// dashboard object

		/******************/
		/* setup language */
		/******************/

		char szLang[3];
		nmCpy (szLang, sizeof (szLang), lpXml+nAttOff, nAttLen);
		CLang::SetLangId (szLang);

		/**********************/
		/* get header element */
		/**********************/

		if (Dash.Parse (&nTagOff, &nTagLen, &nDatOff, &nDatLen))
			return;

		LPSTR lpHea = lpDash + nDatOff;
		int nHeaLen = nDatLen;
		CXml Hea (lpHea, nHeaLen);				// header object

		/********************/
		/* get body element */
		/********************/

		if (Dash.Parse (&nTagOff, &nTagLen, &nDatOff, &nDatLen))
			return;

		LPSTR lpBod = lpDash + nDatOff;
		int nBodLen = nDatLen;
		CXml Bod (lpBod, nBodLen);				// body object

		Bod.Entities();							// decode entities

		/******************************/
		/* get document template path */
		/******************************/

		if (Hea.Parse (&nTagOff, &nTagLen, &nDatOff, &nDatLen))
			return;

		char szDot[MAX_PATH];
		nmCpy (szDot, sizeof (szDot), lpHea+nDatOff, nDatLen);

		/*************************************************************/
		/* get replacements as zero-terminated string pairs in lpSub */
		/*************************************************************/

		int nSub = 0;

		lpSub = (LPSTR) malloc (nSub+1);
		if (lpSub == NULL) {
			CLang::MesBox (IDS_ERRALLOC);
			return;
			}

		*lpSub = '\0';

		while (TRUE) {

			/**************************/
			/* get replacement object */
			/**************************/

			if (Hea.Parse (&nTagOff, &nTagLen, &nDatOff, &nDatLen))
				return;

			if (nmiCmp ("replacement", 11, lpHea+nTagOff, nTagLen))
				break;

			LPSTR lpRep = lpHea + nDatOff;
			int nRepLen = nDatLen;
			CXml Rep (lpRep, nRepLen);			// replacement object

			/**********************/
			/* get string to find */
			/**********************/

			if (Rep.Parse (&nTagOff, &nTagLen, &nDatOff, &nDatLen))
				return;

			lpSub = (LPSTR) realloc (lpSub, nSub+nDatLen+1);
			if (lpSub == NULL) {
				CLang::MesBox (IDS_ERRALLOC);
				return;
				}

			nCpy (lpSub+nSub, lpRep+nDatOff, nDatLen);
			nSub += nDatLen;
			*(lpSub+nSub++) = '\0';

			/************************/
			/* get replacing string */
			/************************/

			if (Rep.Parse (&nTagOff, &nTagLen, &nDatOff, &nDatLen))
				return;

			lpSub = (LPSTR) realloc (lpSub, nSub+nDatLen+2);
			if (lpSub == NULL) {
				CLang::MesBox (IDS_ERRALLOC);
				return;
				}

			nCpy (lpSub+nSub, lpRep+nDatOff, nDatLen);
			nSub += nDatLen;
			*(lpSub+nSub++) = '\0';
			*(lpSub+nSub) = '\0';
			}

		/**********************************************************/
		/* get saveas element and create intermediate directories */
		/**********************************************************/

		char szDoc[MAX_PATH];
		int nDoc = nmCpy (szDoc, sizeof (szDoc), lpHea+nDatOff, nDatLen);
		if (CreatePath (szDoc))
			return;

		/******************/
		/* initialize Ole */
		/******************/

		HRESULT hr = OleInitialize (NULL);		// initialize Ole
		if (hr != S_OK && hr != S_FALSE)
			return;

		bOleInitialized = TRUE;

		/**************/
		/* open Excel */
		/**************/

		CExcel x;								// local CExcel object

		if (x.Open())							// open object
			return;

		/*************/
		/* open Word */
		/*************/

		CWord w;								// local CWord object

		if (w.CloseDocument (szDoc))			// close document in Word running instance
			return;
	
		if (w.Open())							// open object
			return;

		/**********************************************/
		/* create new document from template document */
		/**********************************************/
		{
		LPSTR lpszPath = szDot;					// dot file path
		BOOL bPathIsUrl;						// path is url

		if ((lpszPath = CInet::ReadUrl (lpszPath, &bPathIsUrl)) == 0)
			return;

		w.Add (lpszPath);						// create new document from template

		if (bPathIsUrl) {
			DeleteFile (lpszPath);				// delete temporary file
			free (lpszPath);					// free allocated path
			}
		}

		w.PageView();							// set page view
		w.Zoom (100);							// zoom at 100 %
		w.SaveAs (szDoc);						// save document

		/************************************/
		/* move Word instance to foreground */
		/************************************/
		{
		char szCap[17+MAX_PATH] = "Microsoft Word - ";
		LPSTR lp = rChr(szDoc, '\\')+1;
		int n = lstrlen (lp);
		nmCat (szCap, 17+MAX_PATH, 17, lp, n);
		HWND hWnd = FindWindow (NULL, szCap);
		if (hWnd) SetForegroundWindow (hWnd);
		}
		/************************************/
		/* move Word instance to foreground */
		/************************************/
		{
		char szCap[17+MAX_PATH];
		LPSTR lp = rChr(szDoc, '\\')+1;
		int n = lstrlen (lp);
		n = nmCpy (szCap, 17+MAX_PATH, lp, n);
		nmCat (szCap, 17+MAX_PATH, n, " - Microsoft Word", 17);
		HWND hWnd = FindWindow (NULL, szCap);
		if (hWnd) SetForegroundWindow (hWnd);
		}
		/**************************************************/
		/* do replacements in document header/footer/main */
		/**************************************************/

		w.SeekView (9);							// select header
		{
		for (LPSTR lpszFin = lpSub, lpszRep; *lpszFin; lpszFin = lpszRep + lstrlen (lpszRep) + 1) {
			lpszRep = lpszFin + lstrlen (lpszFin) + 1;
			w.Replace (lpszFin, lpszRep);
			}
		}

		w.SeekView (10);						// select footer
		{
		for (LPSTR lpszFin = lpSub, lpszRep; *lpszFin; lpszFin = lpszRep + lstrlen (lpszRep) + 1) {
			lpszRep = lpszFin + lstrlen (lpszFin) + 1;
			w.Replace (lpszFin, lpszRep);
			}
		}

		w.SeekView (0);							// select main document
		{
		for (LPSTR lpszFin = lpSub, lpszRep; *lpszFin; lpszFin = lpszRep + lstrlen (lpszRep) + 1) {
			lpszRep = lpszFin + lstrlen (lpszFin) + 1;
			w.Replace (lpszFin, lpszRep);
			}
		}

		w.GotoEnd();							// goto end of document

		w.SeekView (9);							// select header
		{
		for (LPSTR lpszFin = lpSub, lpszRep; *lpszFin; lpszFin = lpszRep + lstrlen (lpszRep) + 1) {
			lpszRep = lpszFin + lstrlen (lpszFin) + 1;
			w.Replace (lpszFin, lpszRep);
			}
		}

		w.SeekView (10);						// select footer
		{
		for (LPSTR lpszFin = lpSub, lpszRep; *lpszFin; lpszFin = lpszRep + lstrlen (lpszRep) + 1) {
			lpszRep = lpszFin + lstrlen (lpszFin) + 1;
			w.Replace (lpszFin, lpszRep);
			}
		}

		w.SeekView (0);							// select main document

		w.GotoEnd();							// goto end of document

		free (lpSub);							// free substitution buffer
		lpSub = 0;

		w.PageView();							// set view page
		w.PageBreak();							// write page break
		w.SaveAs (szDoc);						// save document

		/*==================================================================*/
		/* main parsing loop												*/
		/*==================================================================*/

		int nUpdates = 0;

		while (Bod.Parse (&nTagOff, &nTagLen, &nDatOff, &nDatLen, &nAttOff, &nAttLen) == 0) {

			LPSTR lpTag = lpBod + nTagOff;		// pointer to tag
			LPSTR lpDat = lpBod + nDatOff;		// pointer to data
			LPSTR lpAtt = lpBod + nAttOff;		// pointer to attribute

			/*********/
			/* graph */
			/*********/

			if (!nmiCmp (lpTag, nTagLen, "graph", 5)) {

				/******************/
				/* get chart type */
				/******************/

				int nChartType = 0;

				if (!nmiCmp (lpAtt, nAttLen, "barchart", 8))
					nChartType = 51;			// xlColumnClustered
				else if (!nmiCmp (lpAtt, nAttLen, "barchart3d", 10))
					nChartType = 54;			// xl3dColumnClustered
				else if (!nmiCmp (lpAtt, nAttLen, "pie", 3))
					nChartType = 5;				// xlPie
				else if (!nmiCmp (lpAtt, nAttLen, "pie3d", 5))
					nChartType = -4102;			// xl3DPie
				else if (!nmiCmp (lpAtt, nAttLen, "curve", 5))
					nChartType = 65;			// xlLineMarkers
				else if (!nmiCmp (lpAtt, nAttLen, "curve3d", 7))
					nChartType = -4101;			// xl3DLine

				/*********************************/
				/* get chart parameters and data */
				/*********************************/

				CXml Prm (lpDat, nDatLen);		// chart parameters

				int nSizOff, nSizLen;			// graph size
				int nTitOff, nTitLen;			// graph title
				int nXaxOff, nXaxLen;			// X axis title
				int nYaxOff, nYaxLen;			// Y axis title
				int nDa2Off, nDa2Len;			// graph data

				if (Prm.Parse (&nTagOff, &nTagLen, &nSizOff, &nSizLen) ||
					Prm.Parse (&nTagOff, &nTagLen, &nTitOff, &nTitLen) ||
					Prm.Parse (&nTagOff, &nTagLen, &nXaxOff, &nXaxLen) ||
					Prm.Parse (&nTagOff, &nTagLen, &nYaxOff, &nYaxLen) ||
					Prm.Parse (&nTagOff, &nTagLen, &nDa2Off, &nDa2Len))
						return;

				LPSTR lpszSiz = lpDat+nSizOff;
				LPSTR lpszTit = lpDat+nTitOff;
				LPSTR lpszXax = lpDat+nXaxOff;
				LPSTR lpszYax = lpDat+nYaxOff;
				LPSTR lpszDa2 = lpDat+nDa2Off;

				/*********************************/
				/* get Ymin and Ymax if existing */
				/*********************************/

				int nYmiOff, nYmiLen;			// Y axis min
				LPSTR lpszYmi = 0;

				if (!nmiCmp (lpDat+nTagOff, nTagLen, "ymin", 4)) {
					nYmiOff = nDa2Off;
					nYmiLen = nDa2Len;
					lpszYmi = lpszDa2;
					if (*lpszYmi == '<')
						lpszYmi = 0;

					if (Prm.Parse (&nTagOff, &nTagLen, &nDa2Off, &nDa2Len))
						return;

					lpszDa2 = lpDat+nDa2Off;
					}

				int nYmxOff, nYmxLen;			// Y axis min
				LPSTR lpszYmx = 0;

				if (!nmiCmp (lpDat+nTagOff, nTagLen, "ymax", 4)) {
					nYmxOff = nDa2Off;
					nYmxLen = nDa2Len;
					lpszYmx = lpszDa2;
					if (*lpszYmx == '<')
						lpszYmx = 0;

					if (Prm.Parse (&nTagOff, &nTagLen, &nDa2Off, &nDa2Len))
						return;

					lpszDa2 = lpDat+nDa2Off;
					}

				/*********************************/
				/* just write message if no data */
				/*********************************/

				if (nLfs (lpszDa2, nDa2Len) < 2) {

					char szMsg[MAX_PATH];
					int nMsg;

					switch (nChartType) {
					case 51:
					case 54:
						nMsg = CLang::LoadString (IDS_NOBARCHART, szMsg, sizeof (szMsg));
						break;
					case 5:
					case -4102:
						nMsg = CLang::LoadString (IDS_NOPIE, szMsg, sizeof (szMsg));
						break;
					case 65:
					case -4101:
						nMsg = CLang::LoadString (IDS_NOCURVE, szMsg, sizeof (szMsg));
						break;
						}

					if (nMsg)
						w.Normal (szMsg, nMsg);
					}

				/***************************************/
				/* draw chart in excel, insert in word */
				/***************************************/

				else if (nChartType) {

					x.CreateChart (
						lpszTit, nTitLen,		// graph title
						lpszXax, nXaxLen,		// X axis title
						lpszYax, nYaxLen,		// Y axis title
						lpszDa2, nDa2Len,		// graph data
						lpszYmi, nYmiLen,		// Y axis min
						lpszYmx, nYmxLen,		// Y axis max
						nChartType				// chart type
						);

					w.Paste (float (atof (lpszSiz)));	// graph size

					x.DeleteChart();
					}
				}

			/*********/
			/* image */
			/*********/

			else if (!nmiCmp (lpTag, nTagLen, "image", 5)) {

				/******************************/
				/* get image from url or file */
				/******************************/

				BOOL bPathIsUrl;				// path is url
				LPSTR lpszUrl;					// url

				if ((lpszUrl = CInet::ReadUrl (lpDat, nDatLen, &bPathIsUrl)) == 0)
					return;

				w.Image (lpszUrl);				// insert image

				if (bPathIsUrl) {
					DeleteFile (lpszUrl);		// delete temporary file
					free (lpszUrl);				// free allocated path
					}
				}

			/*************/
			/* linebreak */
			/*************/

			else if (!nmiCmp (lpTag, nTagLen, "linebreak", 9)) {
				w.LineBreak();
				}

			/*************/
			/* pagebreak */
			/*************/

			else if (!nmiCmp (lpTag, nTagLen, "pagebreak", 9)) {
				w.PageBreak();
				}

			/*********/
			/* table */
			/*********/

			else if (!nmiCmp (lpTag, nTagLen, "table", 5)) {

				/*********************************/
				/* get table parameters and data */
				/*********************************/

				CXml Table (lpDat, nDatLen);

				int nTa2Off, nTa2Len;			// Tag
				int nSizOff, nSizLen;			// Table size
				int nWidOff, nWidLen;			// Column widthes
				int nAliOff, nAliLen;			// Column alignments
				int nDa2Off, nDa2Len;			// Table data

				if (Table.Parse (&nTa2Off, &nTa2Len, &nSizOff, &nSizLen) ||
					Table.Parse (&nTa2Off, &nTa2Len, &nWidOff, &nWidLen) ||
					Table.Parse (&nTa2Off, &nTa2Len, &nAliOff, &nAliLen) ||
					Table.Parse (&nTa2Off, &nTa2Len, &nDa2Off, &nDa2Len))
						return;

				/***********************/
				/* write table in word */
				/***********************/

				w.Table (
					lpDat+nSizOff, nSizLen,		// Table size
					lpDat+nWidOff, nWidLen,		// Column widthes
					lpDat+nAliOff, nAliLen,		// Column alignments
					lpDat+nDa2Off, nDa2Len		// Table data
					);
				}

			/********/
			/* text */
			/********/

			else if (!nmiCmp (lpTag, nTagLen, "text", 4)) {

				if (!nmiCmp (lpAtt, nAttLen, "normal", 6))
					w.Normal (lpDat, nDatLen);
				else if (!nmiCmp (lpAtt, nAttLen, "pagetitle1", 10))
					w.PageTitle (-2, lpDat, nDatLen);
				else if (!nmiCmp (lpAtt, nAttLen, "title1", 6))
					w.Title (-2, lpDat, nDatLen);
				else if (!nmiCmp (lpAtt, nAttLen, "title2", 6))
					w.Title (-3, lpDat, nDatLen);
				else if (!nmiCmp (lpAtt, nAttLen, "title3", 6))
					w.Title (-4, lpDat, nDatLen);
				else if (!nmiCmp (lpAtt, nAttLen, "title4", 6))
					w.Title (-5, lpDat, nDatLen);
				else if (!nmiCmp (lpAtt, nAttLen, "title5", 6))
					w.Title (-6, lpDat, nDatLen);
				else if (!nmiCmp (lpAtt, nAttLen, "title6", 6))
					w.Title (-7, lpDat, nDatLen);
				}

			/******************************/
			/* save document periodically */
			/******************************/

			if (++nUpdates == 20) {
				nUpdates = 0;
				if (w.GetVersion() == 8)		// if word 97
					w.Save();					// save
				}
			}

		x.Close();								// close CWord object
		w.FieldsUpdate();						// update fields
		w.PageView();							// set page view
		w.Save();								// save document
		w.Close();								// close CWord object
		}

	/*==================================================================*/
	/* free resources section											*/
	/*==================================================================*/

	__finally {

		if (bOleInitialized)
			OleUninitialize();					// uninitialize Ole

		if (lpSub)
			free (lpSub);						// free substitution buffer

		if (lpXml)
			free (lpXml);						// free xml buffer

#ifdef DEBUG
		_CrtMemDumpAllObjectsSince (&state);	// check out memoryleaks
#endif
		}
}
